#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import typing

import octobot_commons.logging as logging
import octobot_commons.singleton as singleton


class ExchangeConfiguration:
    def __init__(self, exchange_manager, matrix_id):
        self.exchange_manager = exchange_manager
        self.exchange_name = exchange_manager.exchange_name
        self.id = exchange_manager.id
        self.matrix_id = matrix_id
        # use only enabled currencies
        # use exchange_manager.exchange_config.all_config_symbol_pairs to get all pairs regardless of whether they
        # are activated in config
        self.symbols = exchange_manager.exchange_config.traded_symbol_pairs
        self.symbols_by_crypto_currencies = exchange_manager.exchange_config.traded_cryptocurrencies
        self.real_time_time_frames = exchange_manager.exchange_config.real_time_time_frames
        self.available_required_time_frames = exchange_manager.exchange_config.available_required_time_frames


class Exchanges(singleton.Singleton):
    def __init__(self):
        self.exchanges = {}

    def add_exchange(self, exchange_manager, matrix_id) -> None:
        if exchange_manager.exchange_name not in self.exchanges:
            self.exchanges[exchange_manager.exchange_name] = {}

        self.exchanges[exchange_manager.exchange_name][exchange_manager.id] = ExchangeConfiguration(exchange_manager,
                                                                                                    matrix_id)

    def get_exchange(self, exchange_name, exchange_manager_id) -> ExchangeConfiguration:
        return self.exchanges[exchange_name][exchange_manager_id]

    def get_all_exchanges(self):
        exchanges_list: list = []
        for exchange_name in self.exchanges.keys():
            exchanges_list += self.get_exchanges_list(exchange_name)
        return exchanges_list

    def get_exchanges(self, exchange_name) -> dict:
        return self.exchanges[exchange_name]

    def get_exchanges_list(self, exchange_name) -> list:
        return list(self.exchanges[exchange_name].values())

    def del_exchange(self, exchange_name, exchange_manager_id, should_warn=True) -> None:
        try:
            self.exchanges[exchange_name].pop(exchange_manager_id, None)

            if not self.exchanges[exchange_name]:
                self.exchanges.pop(exchange_name, None)
        except KeyError:
            if should_warn:
                logging.get_logger(self.__class__.__name__).warning(f"Can't del exchange {exchange_name} "
                                                                    f"with id {exchange_manager_id}")

    def get_exchange_names(self) -> typing.KeysView:
        return self.exchanges.keys()

    def get_exchange_ids(self) -> list:
        return [exchange_id
                for exchange_managers in self.exchanges.values()
                for exchange_id in exchange_managers.keys()]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_trading.exchange_data.ohlcv import candles_manager
from octobot_trading.exchange_data.ohlcv import candles_adapter
from octobot_trading.exchange_data.ohlcv import channel

from octobot_trading.exchange_data.ohlcv.candles_manager import (
    CandlesManager,
)
from octobot_trading.exchange_data.ohlcv.candles_adapter import (
    get_symbol_close_candles,
    get_symbol_open_candles,
    get_symbol_high_candles,
    get_symbol_low_candles,
    get_symbol_volume_candles,
    get_symbol_time_candles,
    get_candle_as_list,
)
from octobot_trading.exchange_data.ohlcv.channel import (
    OHLCVUpdaterSimulator,
    OHLCVProducer,
    OHLCVChannel,
)
from octobot_trading.exchange_data.ohlcv.channel.ohlcv_updater import (
    OHLCVUpdater,
)

__all__ = [
    "CandlesManager",
    "get_symbol_close_candles",
    "get_symbol_open_candles",
    "get_symbol_high_candles",
    "get_symbol_low_candles",
    "get_symbol_volume_candles",
    "get_symbol_time_candles",
    "get_candle_as_list",
    "OHLCVUpdaterSimulator",
    "OHLCVProducer",
    "OHLCVChannel",
    "OHLCVUpdater",
]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import contextlib

import octobot_commons.logging as logging
import octobot_commons.constants as commons_constants
import octobot_commons.tree as commons_tree
import octobot_commons.enums as commons_enums

import octobot_trading.exchange_channel as exchange_channel
import octobot_trading.constants as constants
import octobot_trading.errors as errors
import octobot_trading.personal_data as personal_data
import octobot_trading.util as util


class PortfolioManager(util.Initializable):
    """
    Manage the portfolio and portfolio profitability instances
    """

    def __init__(self, config, trader, exchange_manager):
        super().__init__()
        self.logger = logging.get_logger(self.__class__.__name__)
        self.config, self.trader, self.exchange_manager = config, trader, exchange_manager

        self.portfolio = None
        self.portfolio_profitability = None
        self.portfolio_value_holder = None
        self.historical_portfolio_value_manager = None
        self.reference_market = None
        self._is_initialized_event_set = False

    async def initialize_impl(self):
        """
        Reset the portfolio instance
        """
        self._reset_portfolio()
        if self.historical_portfolio_value_manager is not None:
            await self.historical_portfolio_value_manager.initialize()

    def handle_balance_update(self, balance, is_diff_update=False):
        """
        Handle a balance update request
        :param balance: the new balance
        :param is_diff_update: True when the update is a partial portfolio
        :return: True if the portfolio was updated
        """
        changed = False
        if self.trader.is_enabled and balance is not None:
            changed = self.portfolio.update_portfolio_from_balance(balance, force_replace=not is_diff_update)
        if not self._is_initialized_event_set:
            self._set_initialized_event()
            self._is_initialized_event_set = True
        return changed

    async def handle_balance_update_from_order(self, order, require_exchange_update: bool) -> bool:
        """
        Handle a balance update from an order request
        :param order: the order
        :param require_exchange_update: when True, will sync with exchange portfolio, otherwise will predict the
        portfolio changes using order data (as in trading simulator)
        :return: True if the portfolio was updated
        """
        if self.trader.is_enabled:
            async with self.portfolio_history_update():
                if self.trader.simulate or not require_exchange_update:
                    return self._refresh_simulated_trader_portfolio_from_order(order)
                # on real trading: reload portfolio to ensure portfolio sync
                return await self._refresh_real_trader_portfolio()
        return False

    async def handle_balance_update_from_funding(self, position, funding_rate, require_exchange_update: bool) -> bool:
        """
        Handle a balance update from a funding update
        :param position: the position
        :param funding_rate: the funding rate
        :param require_exchange_update: when True, will sync with exchange portfolio, otherwise will predict the
        portfolio changes using position data (as in trading simulator)
        :return: True if the portfolio was updated
        """
        if self.trader.is_enabled:
            async with self.portfolio_history_update():
                if self.trader.simulate or not require_exchange_update:
                    self.portfolio.update_portfolio_from_funding(position, funding_rate)
                    return True
                # on real trading: reload portfolio to ensure portfolio sync
                return await self._refresh_real_trader_portfolio()
        return False

    async def handle_balance_update_from_withdrawal(self, amount, currency) -> bool:
        """
        Handle a balance update from a withdrawal update
        :param amount: the amount to withdraw
        :param currency: the currency to withdraw
        :return: True if the portfolio was updated
        """
        if self.trader.is_enabled:
            async with self.portfolio_history_update():
                if self.trader.simulate:
                    self.portfolio.update_portfolio_from_withdrawal(amount, currency)
                    return True
                # do not withdraw on real trading
                raise errors.PortfolioOperationError("withdraw is not supported in real trading")
        return False

    def handle_balance_updated(self):
        """
        Handle balance update notification
        :return: True if profitability changed
        """
        return self.portfolio_profitability.update_profitability()

    def get_portfolio_historical_values(self, currency, time_frame, from_timestamp, to_timestamp):
        if self.historical_portfolio_value_manager is None:
            raise errors.NotSupported("historical_portfolio_value_manager has to be set to get historical values")
        historical_values = self.historical_portfolio_value_manager.get_historical_values(
            currency, time_frame, from_timestamp, to_timestamp
        )
        # add/update current portfolio value
        current_historical_time = self.historical_portfolio_value_manager.convert_to_historical_timestamp(
            self.exchange_manager.exchange.get_exchange_current_time(), time_frame
        )
        historical_values[current_historical_time] = self.portfolio_value_holder.portfolio_current_value
        return historical_values

    async def update_historical_portfolio_values(self):
        if self.historical_portfolio_value_manager is None or \
           not self.portfolio_value_holder.current_crypto_currencies_values or \
           self.portfolio_value_holder.initializing_symbol_prices:
            # initializing symbol prices, impossible to get an accurate portfolio value for now
            return
        try:
            # in backtesting, save at the end of the backtesting when calling stop
            await self.historical_portfolio_value_manager.on_new_value(
                self.exchange_manager.exchange.get_exchange_current_time(),
                {
                    self.reference_market: self.portfolio_value_holder.portfolio_current_value
                },
                save_changes=not self.exchange_manager.is_backtesting
            )
        except Exception as e:
            self.logger.exception(e, True, f"Error when saving historical portfolio: {e}")

    @contextlib.asynccontextmanager
    async def portfolio_history_update(self):
        try:
            yield
        finally:
            if self.historical_portfolio_value_manager is not None:
                try:
                    await self.historical_portfolio_value_manager.on_portfolio_update()
                except Exception as err:
                    self.logger.exception(f"Error when updating portfolio history: {err}")

    def handle_profitability_recalculation(self, force_recompute_origin_portfolio):
        """
        Called before PortfolioProfitability's portfolio profitability recalculation
        to ensure portfolio values are available
        :param force_recompute_origin_portfolio: when True, force origin portfolio computation
        """
        self.portfolio_value_holder.handle_profitability_recalculation(force_recompute_origin_portfolio)

    def handle_mark_price_update(self, symbol, mark_price):
        """
        Handle a mark price update notification
        :param symbol: the update symbol
        :param mark_price: the updated mark price in Decimal
        :return: True if profitability changed
        """
        return self.portfolio_profitability. \
            update_profitability(force_recompute_origin_portfolio=self.portfolio_value_holder.
                                 update_origin_crypto_currencies_values(symbol, mark_price))

    async def _refresh_real_trader_portfolio(self) -> bool:
        """
        Call BALANCE_CHANNEL producer to refresh real trader portfolio
        :return: True if the portfolio was updated
        """
        return await exchange_channel.get_chan(constants.BALANCE_CHANNEL,
                                               self.exchange_manager.id).get_internal_producer().\
            refresh_real_trader_portfolio()

    def _reset_portfolio(self):
        """
        Reset the portfolio and portfolio profitability instances
        """
        self.portfolio = personal_data.create_portfolio_from_exchange_manager(self.exchange_manager)
        self._load_portfolio()

        self.reference_market = util.get_reference_market(self.config)
        self.portfolio_value_holder = personal_data.PortfolioValueHolder(self)
        if self.exchange_manager.is_storage_enabled():
            self.historical_portfolio_value_manager = personal_data.HistoricalPortfolioValueManager(self)
        self.portfolio_profitability = personal_data.PortfolioProfitability(self)
        self._is_initialized_event_set = False

    def _refresh_simulated_trader_portfolio_from_order(self, order):
        """
        Handle a balance update from an order request when simulating
        Catch a PortfolioNegativeValueError when calling portfolio update method and returns False if raised
        :param order: the order that should update portfolio
        :return: True if the portfolio was updated
        """
        try:
            if order.is_filled():
                self.portfolio.update_portfolio_from_filled_order(order)
            else:
                self.portfolio.update_portfolio_available(order, is_new_order=False)
            return True
        except errors.PortfolioNegativeValueError as portfolio_negative_value_error:
            self.logger.exception(portfolio_negative_value_error, True,
                                  f"Failed to update portfolio : {portfolio_negative_value_error} "
                                  f"for order {order.to_dict()}")
        return False

    def _load_portfolio(self):
        """
        Load simulated portfolio from config if required
        """
        if self.trader.is_enabled:
            if self.trader.simulate:
                self._set_starting_simulated_portfolio()
            self.logger.info(f"{constants.CURRENT_PORTFOLIO_STRING} {self.portfolio.portfolio}")

    def _set_starting_simulated_portfolio(self):
        """
        Load new portfolio from config settings
        """
        portfolio_amount_dict = personal_data.parse_decimal_config_portfolio(
            self.config[commons_constants.CONFIG_SIMULATOR][commons_constants.CONFIG_STARTING_PORTFOLIO]
        )
        self.handle_balance_update(self.portfolio.get_portfolio_from_amount_dict(portfolio_amount_dict))

    def _set_initialized_event(self):
        commons_tree.EventProvider.instance().trigger_event(
            self.exchange_manager.bot_id, commons_tree.get_exchange_path(
                self.exchange_manager.exchange_name,
                commons_enums.InitializationEventExchangeTopics.BALANCE.value
            )
        )

    async def stop(self):
        if self.historical_portfolio_value_manager is not None:
            await self.historical_portfolio_value_manager.stop()

    def clear(self):
        """
        Clear portfolio manager objects
        """
        self.portfolio_profitability = None
        self.portfolio_value_holder = None
        self.historical_portfolio_value_manager = None



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.logging as logging

from octobot_trading.exchange_data.exchange_symbol_data import ExchangeSymbolData


class ExchangeSymbolsData:
    def __init__(self, exchange_manager):
        self.logger = logging.get_logger(self.__class__.__name__)
        self.exchange_manager = exchange_manager
        self.exchange = exchange_manager.exchange
        self.config = exchange_manager.config
        self.exchange_symbol_data = {}

    def get_exchange_symbol_data(self, symbol, allow_creation=True):
        try:
            return self.exchange_symbol_data[symbol]
        except KeyError as e:
            if allow_creation:
                # warning: should only be called in the async loop thread
                self.exchange_symbol_data[symbol] = ExchangeSymbolData(self.exchange_manager, symbol)
                return self.exchange_symbol_data[symbol]
            raise e



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import copy
import octobot_trading.util as util
import octobot_trading.constants as trading_constants
import octobot_commons.symbols as symbol_util
import octobot_commons.constants as commons_constants

from tests import config


def test_is_trader_enabled(config):
    _test_enabled(config, util.is_trader_enabled, commons_constants.CONFIG_TRADER, False)


def test_is_trader_simulator_enabled(config):
    _test_enabled(config, util.is_trader_simulator_enabled, commons_constants.CONFIG_SIMULATOR, True)


def test_is_trade_history_loading_enabled(config):
    assert util.is_trade_history_loading_enabled(config) is True
    config[commons_constants.CONFIG_TRADER][commons_constants.CONFIG_LOAD_TRADE_HISTORY] = False
    assert util.is_trade_history_loading_enabled(config) is False
    config.pop(commons_constants.CONFIG_TRADER)
    assert util.is_trade_history_loading_enabled(config) is True
    assert util.is_trade_history_loading_enabled(config, False) is False


def test_is_currency_enabled(config):
    config[commons_constants.CONFIG_CRYPTO_CURRENCIES]["Bitcoin"][commons_constants.CONFIG_ENABLED_OPTION] = True
    assert util.is_currency_enabled(config, "Bitcoin", True) is True
    assert util.is_currency_enabled(config, "Bitcoin", False) is True
    config[commons_constants.CONFIG_CRYPTO_CURRENCIES]["Bitcoin"][commons_constants.CONFIG_ENABLED_OPTION] = False
    assert util.is_currency_enabled(config, "Bitcoin", True) is False
    config[commons_constants.CONFIG_CRYPTO_CURRENCIES]["Bitcoin"].pop(commons_constants.CONFIG_ENABLED_OPTION, None)
    assert util.is_currency_enabled(config, "Bitcoin", True) is True
    assert util.is_currency_enabled(config, "Bitcoin", False) is False
    assert util.is_currency_enabled(config, "Hello", False) is False


def test_get_symbols(config):
    assert util.get_symbols(config, True) == FULL_PAIRS_LIST
    # with wildcard currency
    config[commons_constants.CONFIG_CRYPTO_CURRENCIES]["Bitcoin"][commons_constants.CONFIG_CRYPTO_PAIRS] = \
        commons_constants.CONFIG_WILDCARD
    list_without_bitcoin = _filter_by_base(FULL_PAIRS_LIST, "BTC")
    assert util.get_symbols(config, True) == list_without_bitcoin

    # with disabled currency
    config[commons_constants.CONFIG_CRYPTO_CURRENCIES]["Bitcoin"][commons_constants.CONFIG_CRYPTO_PAIRS] = [
        'BTC/USDT',
        'BTC/EUR',
        'BTC/USDC'
    ]
    config[commons_constants.CONFIG_CRYPTO_CURRENCIES]["Bitcoin"][commons_constants.CONFIG_ENABLED_OPTION] = False
    list_without_bitcoin = _filter_by_base(FULL_PAIRS_LIST, "BTC")
    assert util.get_symbols(config, True) == list_without_bitcoin
    assert util.get_symbols(config, False) == FULL_PAIRS_LIST

    # with empty pairs
    config[commons_constants.CONFIG_CRYPTO_CURRENCIES]["Ethereum"][commons_constants.CONFIG_CRYPTO_PAIRS] = []
    assert util.get_symbols(config, False) == _filter_by_base(FULL_PAIRS_LIST, "ETH")

    # with broken config
    config[commons_constants.CONFIG_CRYPTO_CURRENCIES] = []
    assert util.get_symbols(config, False) == []
    config.pop(commons_constants.CONFIG_CRYPTO_CURRENCIES, None)
    assert util.get_symbols(config, False) == []


def test_get_all_currencies(config):
    symbols = set()
    for pair in FULL_PAIRS_LIST:
        symbols.update(symbol_util.parse_symbol(pair).base_and_quote())
    assert util.get_all_currencies(config) == symbols
    assert util.get_all_currencies(config, enabled_only=True) == symbols

    # with disabled currency
    config[commons_constants.CONFIG_CRYPTO_CURRENCIES]["Bitcoin"][commons_constants.CONFIG_ENABLED_OPTION] = False
    assert util.get_all_currencies(config) == symbols
    symbols_without_bitcoin_symbols = copy.copy(symbols)
    symbols_without_bitcoin_symbols.remove("EUR")
    symbols_without_bitcoin_symbols.remove("USDC")
    assert util.get_all_currencies(config, enabled_only=True) == symbols_without_bitcoin_symbols
    assert util.get_all_currencies(config, enabled_only=False) == symbols


def test_get_pairs(config):
    assert util.get_pairs(config, "BTC") == _select_by_base_or_quote(FULL_PAIRS_LIST, "BTC")
    assert util.get_pairs(config, "ICX") == ["ICX/BTC"]

    # with disabled currency
    config[commons_constants.CONFIG_CRYPTO_CURRENCIES]["Bitcoin"][commons_constants.CONFIG_ENABLED_OPTION] = False
    assert util.get_pairs(config, "BTC", enabled_only=False) == _select_by_base_or_quote(FULL_PAIRS_LIST, "BTC")
    assert util.get_pairs(config, "BTC", enabled_only=True) != _select_by_base_or_quote(FULL_PAIRS_LIST, "BTC")
    assert util.get_pairs(config, "BTC", enabled_only=True) == _filter_by_base(_select_by_base_or_quote(FULL_PAIRS_LIST,
                                                                                                        "BTC"),
                                                                               "BTC")


def test_get_market_pair(config):
    config[commons_constants.CONFIG_TRADING][commons_constants.CONFIG_TRADER_REFERENCE_MARKET] = "BTC"
    assert util.get_market_pair(config, "ETH") == ("", False)
    assert util.get_market_pair(config, "ADA") == ("ADA/BTC", False)
    assert util.get_market_pair(config, "USDT") == ("BTC/USDT", True)
    assert util.get_market_pair(config, "BTC") == ("", False)

    # with disabled currency
    config[commons_constants.CONFIG_CRYPTO_CURRENCIES]["Cardano"][commons_constants.CONFIG_ENABLED_OPTION] = False
    assert util.get_market_pair(config, "ADA", enabled_only=False) == ("ADA/BTC", False)
    assert util.get_market_pair(config, "ADA", enabled_only=True) == ("", False)


def test_get_reference_market(config):
    config[commons_constants.CONFIG_TRADING][commons_constants.CONFIG_TRADER_REFERENCE_MARKET] = "BTC"
    assert util.get_reference_market(config) == "BTC"
    config[commons_constants.CONFIG_TRADING][commons_constants.CONFIG_TRADER_REFERENCE_MARKET] = "BTC1"
    assert util.get_reference_market(config) == "BTC1"
    config[commons_constants.CONFIG_TRADING].pop(commons_constants.CONFIG_TRADER_REFERENCE_MARKET, None)
    assert util.get_reference_market(config) == trading_constants.DEFAULT_REFERENCE_MARKET


def test_get_traded_pairs_by_currency(config):
    assert util.get_traded_pairs_by_currency(config) == FULL_PAIRS_BY_CRYPTO_DICT
    # with wildcard currency
    config[commons_constants.CONFIG_CRYPTO_CURRENCIES]["Bitcoin"][commons_constants.CONFIG_CRYPTO_PAIRS] = \
        commons_constants.CONFIG_WILDCARD
    dict_without_bitcoin = _replace_value_by_key(FULL_PAIRS_BY_CRYPTO_DICT, "Bitcoin", commons_constants.CONFIG_WILDCARD)
    assert util.get_traded_pairs_by_currency(config) == dict_without_bitcoin

    # with disabled currency
    config[commons_constants.CONFIG_CRYPTO_CURRENCIES]["Bitcoin"][commons_constants.CONFIG_CRYPTO_PAIRS] = [
        'BTC/USDT',
        'BTC/EUR',
        'BTC/USDC'
    ]
    config[commons_constants.CONFIG_CRYPTO_CURRENCIES]["Bitcoin"][commons_constants.CONFIG_ENABLED_OPTION] = False
    dict_without_bitcoin = _filter_by_key(FULL_PAIRS_BY_CRYPTO_DICT, "Bitcoin")
    assert util.get_traded_pairs_by_currency(config) == dict_without_bitcoin

    # with empty pairs
    config[commons_constants.CONFIG_CRYPTO_CURRENCIES]["Bitcoin"][commons_constants.CONFIG_CRYPTO_PAIRS] = []
    assert util.get_traded_pairs_by_currency(config) == _filter_by_key(FULL_PAIRS_BY_CRYPTO_DICT, "Bitcoin")


FULL_PAIRS_LIST = [
    'BTC/USDT',
    'BTC/EUR',
    'BTC/USDC',
    'NEO/BTC',
    'ETH/USDT',
    'ICX/BTC',
    'VEN/BTC',
    'XRB/BTC',
    'ADA/BTC',
    'ONT/BTC',
    'XLM/BTC',
    'POWR/BTC',
    'ETC/BTC',
    'WAX/BTC',
    'XRP/BTC',
    'XVG/BTC'
]

FULL_PAIRS_BY_CRYPTO_DICT = {
    'Bitcoin': ['BTC/USDT', 'BTC/EUR', 'BTC/USDC'],
    'Neo': ['NEO/BTC'],
    'Ethereum': ['ETH/USDT'],
    'Icon': ['ICX/BTC'],
    'VeChain': ['VEN/BTC'],
    'Nano': ['XRB/BTC'],
    'Cardano': ['ADA/BTC'],
    'Ontology': ['ONT/BTC'],
    'Stellar': ['XLM/BTC'],
    'Power Ledger': ['POWR/BTC'],
    'Ethereum Classic': ['ETC/BTC'],
    'WAX': ['WAX/BTC'],
    'XRP': ['XRP/BTC'],
    'Verge': ['XVG/BTC']
}


def _test_enabled(config, func, config_key, current_val):
    assert func(config) is current_val
    config[config_key][commons_constants.CONFIG_ENABLED_OPTION] = not current_val
    assert func(config) is not current_val
    config[config_key].pop(commons_constants.CONFIG_ENABLED_OPTION, None)
    assert func(config) is False
    config.pop(config_key, None)
    assert func(config) is False


def _filter_by_base(pairs, filtered_base):
    return [s for s in pairs if symbol_util.parse_symbol(s).base != filtered_base]


def _filter_by_key(pairs_by_cypto, filtered_key):
    return {
        crypto: pairs
        for crypto, pairs in pairs_by_cypto.items()
        if crypto != filtered_key
    }


def _replace_value_by_key(pairs_by_cypto, filtered_key, replaced_value):
    return {
        crypto: pairs if crypto != filtered_key else replaced_value
        for crypto, pairs in pairs_by_cypto.items()
    }


def _select_by_base_or_quote(pairs, base_or_quote):
    return [s for s in pairs if base_or_quote in symbol_util.parse_symbol(s).base_and_quote()]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest
import pytest_asyncio
import mock
import ccxt.async_support


import octobot_trading.exchanges as exchanges
import octobot_trading.enums as enums

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


@pytest_asyncio.fixture
async def basic_exchange_wrapper():
    async with exchanges.temporary_exchange_wrapper("binance", enums.ExchangeWrapperLibs.ASYNC_CCXT) as wrapper:
        return wrapper


async def test_constructor(basic_exchange_wrapper):
    assert isinstance(basic_exchange_wrapper.exchange, ccxt.async_support.binance)
    async with exchanges.temporary_exchange_wrapper("ftx", enums.ExchangeWrapperLibs.CCXT) as wrapper:
        assert isinstance(wrapper.exchange, ccxt.ftx)


async def test_temporary_exchange_wrapper():
    with mock.patch.object(ccxt.async_support.ftx, "close", mock.AsyncMock()) as close_mock:
        with pytest.raises(ZeroDivisionError):
            async with exchanges.temporary_exchange_wrapper("ftx", enums.ExchangeWrapperLibs.ASYNC_CCXT) as wrapper:
                assert isinstance(wrapper.exchange, ccxt.async_support.ftx)
                close_mock.assert_not_called()
                1/0
        close_mock.assert_called_once()


async def test_unsupported_lib():
    with pytest.raises(NotImplementedError):
        exchanges.BasicExchangeWrapper("binance", "plop")


async def test_get_available_time_frames(basic_exchange_wrapper):
    assert len(await basic_exchange_wrapper.get_available_time_frames()) > 10
    basic_exchange_wrapper.exchange.timeframes = ["1m"]
    assert await basic_exchange_wrapper.get_available_time_frames() == ["1m"]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os
from math import nan

from octobot_trading.enums import ExchangeConstantsMarketStatusColumns as Ecmsc, \
    ExchangeConstantsMarketStatusInfoColumns as Ecmsic
from octobot_trading.exchanges.util.exchange_market_status_fixer import ExchangeMarketStatusFixer


class TestExchangeMarketStatusFixer:

    @staticmethod
    def _get_limits(amount_min=None, amount_max=None,
                    price_min=None, price_max=None,
                    cost_min=None, cost_max=None):
        return {
            Ecmsc.LIMITS_AMOUNT.value: {
                Ecmsc.LIMITS_AMOUNT_MIN.value: amount_min,
                Ecmsc.LIMITS_AMOUNT_MAX.value: amount_max,
            },
            Ecmsc.LIMITS_PRICE.value: {
                Ecmsc.LIMITS_PRICE_MIN.value: price_min,
                Ecmsc.LIMITS_PRICE_MAX.value: price_max,
            },
            Ecmsc.LIMITS_COST.value: {
                Ecmsc.LIMITS_COST_MIN.value: cost_min,
                Ecmsc.LIMITS_COST_MAX.value: cost_max,
            },
        }

    @staticmethod
    def _get_precision(precision_amount=None, precision_cost=None, precision_price=None):
        return {
            Ecmsc.PRECISION_AMOUNT.value: precision_amount,
            Ecmsc.PRECISION_COST.value: precision_cost,
            Ecmsc.PRECISION_PRICE.value: precision_price,
        }

    # Global
    def test_exchange_market_status_fixer_without_precision_cost_and_amount_with_price(self):
        current_price = 4564.1458

        ms = {
            Ecmsc.PRECISION.value: self._get_precision(1, 0, 1),
            Ecmsc.LIMITS.value: self._get_limits(None, nan, 0.05, 1e4, nan, None)
        }

        assert ExchangeMarketStatusFixer(ms, price_example=current_price).market_status == {
            Ecmsc.PRECISION.value: self._get_precision(1, 0, 1),
            Ecmsc.LIMITS.value: self._get_limits(0.002190990480628382, 219.0990480628382,
                                                 current_price / 1000, current_price * 1000,
                                                 0.010000000000000012, 1000000000.0000011)
        }

        ms = {
            Ecmsc.PRECISION.value: self._get_precision(nan, None, nan),
            Ecmsc.LIMITS.value: self._get_limits(None, nan, 0.05, 1e4, nan, None)
        }

        assert ExchangeMarketStatusFixer(ms, price_example=current_price).market_status == {
            Ecmsc.PRECISION.value: self._get_precision(4, 4, 4),
            Ecmsc.LIMITS.value: self._get_limits(0.002190990480628382, 219.0990480628382,
                                                 current_price / 1000, current_price * 1000,
                                                 0.010000000000000012, 1000000000.0000011)
        }

        current_price = 1.56e-6
        ms = {
            Ecmsc.PRECISION.value: self._get_precision(nan, None, nan),
            Ecmsc.LIMITS.value: self._get_limits(None, nan, None, None, nan, None)
        }

        assert ExchangeMarketStatusFixer(ms, price_example=current_price).market_status == {
            Ecmsc.PRECISION.value: self._get_precision(8, 8, 8),
            Ecmsc.LIMITS.value: self._get_limits(641.0256410256403, 6410256.4102564035,
                                                 current_price / 1000, current_price * 1000,
                                                 9.999999999999991e-07, 9999.99999999999)
        }

        current_price = 1.5678999
        ms = {
            Ecmsc.PRECISION.value: self._get_precision(nan, None, nan),
            Ecmsc.LIMITS.value: self._get_limits(None, nan, 0, 0, nan, None)
        }

        assert ExchangeMarketStatusFixer(ms, price_example=current_price).market_status == {
            Ecmsc.PRECISION.value: self._get_precision(7, 7, 7),
            Ecmsc.LIMITS.value: self._get_limits(6.37795818470299, 637795.8184702988,
                                                 current_price / 1000, current_price * 1000,
                                                 0.01, 999999999.9999996)
        }

        current_price = 25.87257
        ms = {
            Ecmsc.PRECISION.value: self._get_precision(),
            Ecmsc.LIMITS.value: self._get_limits(None, None, 0, 0, None, None)
        }

        assert ExchangeMarketStatusFixer(ms, price_example=current_price).market_status == {
            Ecmsc.PRECISION.value: self._get_precision(5, 5, 5),
            Ecmsc.LIMITS.value: self._get_limits(0.3865097282566056, 38650.97282566052,
                                                 current_price / 1000, current_price * 1000,
                                                 0.010000000000000007, 999999999.9999996)
        }

        current_price = 200.555
        ms = {
            Ecmsc.PRECISION.value: self._get_precision(nan, nan, nan),
            Ecmsc.LIMITS.value: self._get_limits(nan, nan, 3, 3, nan, nan)
        }

        assert ExchangeMarketStatusFixer(ms, price_example=current_price).market_status == {
            Ecmsc.PRECISION.value: self._get_precision(3, 3, 3),
            Ecmsc.LIMITS.value: self._get_limits(0.04986163396574511, 4986.163396574511,
                                                 current_price / 1000, current_price * 1000,
                                                 0.01000000000000001, 1000000000.0000011)
        }

    def test_exchange_market_status_fixer_without_market_status(self):
        assert ExchangeMarketStatusFixer({}).market_status == {
            Ecmsc.PRECISION.value: self._get_precision(None, None, None),
            Ecmsc.LIMITS.value: self._get_limits(None, None, None, None, 0, None)
        }

    def test_exchange_market_status_fixer_with_str_instead_of_floats(self):
        ms = {
            Ecmsc.PRECISION.value: self._get_precision("5", "5.777772", None),
            Ecmsc.LIMITS.value: self._get_limits("0.05", "1e4", "0.01", "1e4", "3.3", "11111111111")
        }

        assert ExchangeMarketStatusFixer(ms).market_status == {
            Ecmsc.PRECISION.value: self._get_precision(5, 5.777772, None),
            Ecmsc.LIMITS.value: self._get_limits(0.05, 1e4, 0.01, 1e4, 3.3, 11111111111)
        }

        assert ExchangeMarketStatusFixer(
            {Ecmsc.LIMITS.value: self._get_limits("0.05", "1e4", "plop", nan, "3.3", "11111111111")}
        ).market_status == {
            Ecmsc.PRECISION.value: self._get_precision(None, None, None),
            # replace "plop" and nan with computed numbers
            Ecmsc.LIMITS.value: self._get_limits(0.05, 1e4, 65.99999999999999, 1111111.1111, 3.3, 11111111111)
        }

        # missing and added keys
        limits = self._get_limits("0.05", "1e4", "plop", nan, "3.3", "11111111111")
        limits["plop"] = {"a": "1"}
        limits[Ecmsc.LIMITS_AMOUNT.value]["plop"] = "2"
        limits[Ecmsc.LIMITS_AMOUNT.value].pop(Ecmsc.LIMITS_AMOUNT_MIN.value)

        expected_limits = self._get_limits(0.05, 1e4, "plop", 1111111.1111, 3.3, 11111111111)
        expected_limits[Ecmsc.LIMITS_AMOUNT.value].pop(Ecmsc.LIMITS_AMOUNT_MIN.value)
        expected_limits[Ecmsc.LIMITS_AMOUNT.value]["plop"] = "2"
        expected_limits["plop"] = {"a": "1"}

        assert ExchangeMarketStatusFixer(
            {Ecmsc.LIMITS.value: limits}
        ).market_status == {
            Ecmsc.PRECISION.value: self._get_precision(None, None, None),
            # replace "plop" and nan with computed numbers
            Ecmsc.LIMITS.value: expected_limits
        }

    def test_exchange_market_status_fixer_without_cost(self):
        ms = {
            Ecmsc.PRECISION.value: self._get_precision(5, 5, 5),
            Ecmsc.LIMITS.value: self._get_limits(0.05, 1e4, 0.01, 1e4, None, None)
        }

        assert ExchangeMarketStatusFixer(ms).market_status == {
            Ecmsc.PRECISION.value: self._get_precision(5, 5, 5),
            Ecmsc.LIMITS.value: self._get_limits(0.05, 1e4, 0.01, 1e4, 0.01 * 0.05, 1e4 * 1e4)
        }

    def test_exchange_market_status_fixer_without_price(self):
        ms = {
            Ecmsc.PRECISION.value: self._get_precision(5, 5, 5),
            Ecmsc.LIMITS.value: self._get_limits(0.01, 1e3, nan, nan, 0.05, 1e5)
        }

        assert ExchangeMarketStatusFixer(ms).market_status == {
            Ecmsc.PRECISION.value: self._get_precision(5, 5, 5),
            Ecmsc.LIMITS.value: self._get_limits(0.01, 1e3, 0.05 / 0.01, 1e5 / 1e3, 0.05, 1e5)
        }

    def test_exchange_market_status_fixer_without_price_amount(self):
        ms = {
            Ecmsc.PRECISION.value: self._get_precision(5, 5, 5),
            Ecmsc.LIMITS.value: self._get_limits(nan, None, 0.03, 1e4, 0.05, 1e7)
        }

        assert ExchangeMarketStatusFixer(ms).market_status == {
            Ecmsc.PRECISION.value: self._get_precision(5, 5, 5),
            Ecmsc.LIMITS.value: self._get_limits(0.05 / 0.03, 1e7 / 1e4, 0.03, 1e4, 0.05, 1e7)
        }

    # Limits
    def test_fix_market_status_limits(self):
        if not os.getenv('CYTHON_IGNORE'):
            from octobot_trading.exchanges.util.exchange_market_status_fixer import check_market_status_limits
            assert not check_market_status_limits(self._get_limits(4, None, None, 1000, 56, 45))
            assert not check_market_status_limits(self._get_limits(9, None, 5066, 1000, 56, nan))
            assert not check_market_status_limits(self._get_limits(8, nan, 789, 1000, nan, 45))
            assert not check_market_status_limits(self._get_limits(0, 0, 789, 1000, 0, 45))
            assert check_market_status_limits(self._get_limits(12, 2752783, 242, 1000, 56, 45))

    def test_check_market_status_values(self):
        if not os.getenv('CYTHON_IGNORE'):
            from octobot_trading.exchanges.util.exchange_market_status_fixer import check_market_status_values
            assert not check_market_status_values([78272, None, None, 5e-10, 100, 0.12])
            assert not check_market_status_values([78272, None, nan, 5e-10, 100, 0.12])
            assert not check_market_status_values([78272, nan, nan, 5e-10, 100, 0.12])
            assert not check_market_status_values([78272, 0, 0, 5e-10, 100, 0.12])
            assert check_market_status_values([17, 78272, 79, 5e-10, 145, 100])

    def test_fix_market_status_limits_with_price(self):
        emsf = ExchangeMarketStatusFixer({}, 98765)
        if not os.getenv('CYTHON_IGNORE'):
            emsf._fix_market_status_limits_with_price()
            assert emsf.market_status[Ecmsc.LIMITS.value] == self._get_limits(0.00010125044297068805,
                                                                              10.125044297068806,
                                                                              98.765, 98765000,
                                                                              0.010000000000000005,
                                                                              1000000000.0000006)

            emsf = ExchangeMarketStatusFixer({}, 0.00123456)
            emsf._fix_market_status_limits_with_price()
            assert emsf.market_status[Ecmsc.LIMITS.value] == self._get_limits(0.8100051840331779, 8100.051840331779,
                                                                              1.23456e-06, 1.23456,
                                                                              1.0000000000000002e-06,
                                                                              10000.000000000002)

            emsf = ExchangeMarketStatusFixer({}, 0.0000012)
            emsf._fix_market_status_limits_with_price()
            assert emsf.market_status[Ecmsc.LIMITS.value] == self._get_limits(833.3333333333311, 8333333.333333311,
                                                                              1.2e-9, 0.0012,
                                                                              9.999999999999974e-07,
                                                                              9999.999999999973)

            emsf = ExchangeMarketStatusFixer({}, 0.000999)
            emsf._fix_market_status_limits_with_price()
            assert emsf.market_status[Ecmsc.LIMITS.value] == self._get_limits(1.001001001001,
                                                                              10010.010010009999,
                                                                              9.99e-07, 0.9990000000000001,
                                                                              9.999999999999991e-07,
                                                                              9999.99999999999)

    # Precision
    def test_get_price_precision(self):
        if not os.getenv('CYTHON_IGNORE'):
            assert ExchangeMarketStatusFixer({}, 10.5555)._get_price_precision() == 4
            assert ExchangeMarketStatusFixer({}, 1014578587.5)._get_price_precision() == 1
            assert ExchangeMarketStatusFixer({}, 1.00000000055)._get_price_precision() == 11
            assert ExchangeMarketStatusFixer({}, 1)._get_price_precision() == 0

    def test_fix_market_status_precision_with_price(self):
        if not os.getenv('CYTHON_IGNORE'):
            emsf = ExchangeMarketStatusFixer({}, 10234.55)
            emsf._fix_market_status_precision_with_price()
            assert emsf.market_status[Ecmsc.PRECISION.value] == self._get_precision(2, 2, 2)

            emsf = ExchangeMarketStatusFixer({}, 10234)
            emsf._fix_market_status_precision_with_price()
            assert emsf.market_status[Ecmsc.PRECISION.value] == self._get_precision(0, 0, 0)

            emsf = ExchangeMarketStatusFixer({}, 10.234140561412567)
            emsf._fix_market_status_precision_with_price()
            assert emsf.market_status[Ecmsc.PRECISION.value] == self._get_precision(15, 15, 15)

    # Specific
    def test_fix_market_status_precision_with_specific(self):
        pass

    def test_fix_market_status_limits_with_specific(self):
        # binance specific
        market_status = {
            Ecmsc.INFO.value: {
                Ecmsic.FILTERS.value: [
                    {
                        Ecmsic.FILTER_TYPE.value: Ecmsic.PRICE_FILTER.value,
                        Ecmsic.MAX_PRICE.value: 123456789,
                        Ecmsic.MIN_PRICE.value: 0.1234567
                    },
                    {
                        Ecmsic.FILTER_TYPE.value: Ecmsic.LOT_SIZE.value,
                        Ecmsic.MAX_QTY.value: 9e11,
                        Ecmsic.MIN_QTY.value: 5e-11
                    }
                ]
            }
        }
        emsf = ExchangeMarketStatusFixer(market_status)
        if not os.getenv('CYTHON_IGNORE'):
            emsf._fix_market_status_limits_with_specific()
            assert emsf.market_status[Ecmsc.LIMITS.value] == self._get_limits(5e-11, 9e11,
                                                                              0.1234567, 123456789,
                                                                              5e-11 * 0.1234567, 9e11 * 123456789)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.errors as errors
import octobot_commons.logging as logging
import octobot_commons.tentacles_management as class_inspector

import octobot_tentacles_manager.api as configurator

import octobot_trading.modes as modes


def get_activated_trading_mode(tentacles_setup_config) -> modes.AbstractTradingMode.__class__:
    if tentacles_setup_config is not None:
        try:
            trading_modes = [tentacle_class
                             for tentacle_class in configurator.get_activated_tentacles(tentacles_setup_config)
                             if class_inspector.get_deep_class_from_parent_subclasses(tentacle_class,
                                                                                      modes.AbstractTradingMode)]

            if len(trading_modes) > 1:
                raise errors.ConfigTradingError(
                    f"More than one activated trading mode found in your tentacle configuration, "
                    f"please activate only one")

            elif trading_modes:
                trading_mode_class = class_inspector.get_deep_class_from_parent_subclasses(trading_modes[0],
                                                                                           modes.AbstractTradingMode)

                if trading_mode_class is not None:
                    return trading_mode_class
        except ModuleNotFoundError as e:
            logging.get_logger("get_activated_trading_mode").error(f"Error when loading "
                                                                   f"the activated trading mode: {e}")

    raise errors.ConfigTradingError(f"Please ensure your tentacles configuration file is valid and "
                                    f"at least one trading mode is activated")



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.exchanges.abstract_exchange as abstract_exchange
import octobot_trading.exchanges.connectors as exchange_connectors


# TODO
class RestExchange(abstract_exchange.AbstractExchange):
    """
    RestExchange is only calling the right exchange connector and should be used for each exchange
    request regardless of the trading type (spot / future / etc)
    """
    DEFAULT_CONNECTOR_CLASS = exchange_connectors.CCXTExchange

    def __init__(self, config, exchange_manager, connector_class=DEFAULT_CONNECTOR_CLASS):
        super().__init__(config, exchange_manager)
        self.connector = connector_class(
            config,
            exchange_manager,
            additional_ccxt_config=self.get_additional_connector_config()  # move to connector
        )
        # commented for pylint
        # self.connector.client.options['defaultType'] = self.get_default_type() # move to connector

    async def initialize_impl(self):
        await self.connector.initialize()
        self.symbols = self.connector.symbols
        self.time_frames = self.connector.time_frames

    async def stop(self) -> None:
        await self.connector.stop()
        self.exchange_manager = None



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.enums as enums
import octobot_trading.personal_data.orders.order_state as order_state


class PendingCreationOrderState(order_state.OrderState):

    def __init__(self, order, is_from_exchange_data):
        super().__init__(order, is_from_exchange_data)
        self.state = enums.States.PENDING_CREATION

    def is_created(self) -> bool:
        """
        :return: True if the Order is created
        """
        return False

    def is_open(self) -> bool:
        """
        :return: True if the Order is considered as open
        """
        return False

    async def update(self) -> None:
        # nothing to do as no actual order exists yet
        pass

    async def on_refresh_successful(self):
        # nothing to do as no actual order exists yet
        pass

    async def terminate(self):
        """
        Should wait for being replaced by an OpenOrderState or be cancelled
        """
        self.log_event_message(enums.StatesMessages.PENDING_CREATION)

    def is_pending(self) -> bool:
        return False



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import octobot_trading.modes as modes


def get_trading_modes(exchange_manager) -> list:
    return exchange_manager.trading_modes


def get_trading_mode_symbol(trading_mode) -> list:
    return trading_mode.symbol


def get_trading_mode_current_state(trading_mode) -> tuple:
    return trading_mode.get_current_state()


def get_activated_trading_mode(tentacles_setup_config) -> modes.AbstractTradingMode.__class__:
    return modes.get_activated_trading_mode(tentacles_setup_config)


async def create_trading_modes(config: dict,
                               exchange_manager: object,
                               trading_mode_class: modes.AbstractTradingMode.__class__,
                               bot_id: str) -> list:
    return await modes.create_trading_modes(config=config,
                                            exchange_manager=exchange_manager,
                                            trading_mode_class=trading_mode_class,
                                            bot_id=bot_id)


async def create_trading_mode(trading_mode_class: modes.AbstractTradingMode.__class__,
                              config: dict,
                              exchange_manager: object,
                              cryptocurrency: str = None,
                              symbol: str = None,
                              time_frame: object = None,
                              bot_id: str = None) -> modes.AbstractTradingMode:
    return await modes.create_trading_mode(trading_mode_class=trading_mode_class,
                                           config=config,
                                           exchange_manager=exchange_manager,
                                           cryptocurrency=cryptocurrency,
                                           symbol=symbol,
                                           time_frame=time_frame,
                                           bot_id=bot_id)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library
import octobot_commons.enums as commons_enums
import octobot_commons.constants as commons_constants
import octobot_commons.databases as commons_databases

import octobot_trading.storage.abstract_storage as abstract_storage


class PortfolioStorage(abstract_storage.AbstractStorage):
    IS_LIVE_CONSUMER = False
    IS_HISTORICAL = False

    async def save_historical_portfolio_value(self, metadata, historical_portfolio_value):
        if not self.enabled:
            return
        portfolio_db = self.get_db()
        # replace the whole table to ensure consistency
        await portfolio_db.upsert(commons_enums.RunDatabases.METADATA.value, metadata, None, uuid=1)
        await portfolio_db.replace_all(
            commons_enums.RunDatabases.HISTORICAL_PORTFOLIO_VALUE.value,
            historical_portfolio_value,
            cache=False
        )
        if not self.exchange_manager.is_backtesting:
            # in live move, flush database as soon as an update is provided
            await portfolio_db.flush()

    def get_db(self):
        return commons_databases.RunDatabasesProvider.instance().get_historical_portfolio_value_db(
            self.exchange_manager.bot_id,
            self.exchange_manager.exchange_name,
            self.get_portfolio_type_suffix()
        )

    def get_portfolio_type_suffix(self):
        suffix = ""
        if self.exchange_manager.is_future:
            suffix = f"{suffix}_{commons_constants.CONFIG_EXCHANGE_FUTURE}"
        elif self.exchange_manager.is_margin:
            suffix = f"{suffix}_{commons_constants.CONFIG_EXCHANGE_MARGIN}"
        else:
            suffix = f"{suffix}_{commons_constants.CONFIG_EXCHANGE_SPOT}"
        if self.exchange_manager.is_sandboxed:
            suffix = f"{suffix}_{commons_constants.CONFIG_EXCHANGE_SANDBOXED}"
        if self.exchange_manager.is_trader_simulated:
            suffix = f"{suffix}_{commons_constants.CONFIG_SIMULATOR}"
        return suffix



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport octobot_trading.exchanges.abstract_websocket_exchange as abstract_websocket

cdef class CryptofeedWebsocketConnector(abstract_websocket.AbstractWebsocketExchange):
    cdef public dict callback_by_feed
    cdef public dict callbacks

    cdef public list filtered_pairs
    cdef public list watched_pairs

    cdef public object min_timeframe
    cdef public object candle_callback
    cdef public object cryptofeed_exchange
    cdef public object client_logger
    cdef public object client_config

    cdef public object local_loop
    cdef public bint is_websocket_restarting
    cdef dict _previous_open_candles

    cpdef void start(self)
    cpdef void _set_async_callbacks(self)

    # warning, as cdef functions, those can't be called from an exchange ws tentacle. Migrate to cpdef if a
    # call is necessary
    cdef void _create_client(self)
    cdef void _create_client_config(self)
    cdef dict _get_credentials_config(self)
    cdef void _init_client(self)
    cdef void _start_client(self, bint should_create_loop=*)
    cdef bint _should_use_authenticated_feeds(self)
    cdef dict _get_feed_default_kwargs(self)
    cdef void _subscribe_feeds(self)
    cdef void _add_pair(self, str pair, bint watching_only)
    cdef void _add_time_frame(self, list filtered_timeframes, object time_frame, bint log_on_error)
    cdef void _subscribe_feeds(self)
    cdef bint _is_supported_channel(self, str channel)
    cdef bint _is_supported_pair(self, pair)
    cdef bint _is_supported_time_frame(self, object time_frame)
    cdef bint _is_pair_independent_feed(self, feed)
    cdef object _subscribe_feed(self, list channels, dict callbacks, list symbols=*, str candle_interval=*)
    cdef void _filter_exchange_pairs_and_timeframes(self)
    cdef void _filter_exchange_symbols(self)
    cdef void _filter_exchange_time_frames(self)
    cdef void _subscribe_candle_feed(self)
    cdef void _subscribe_channels_feeds(self)
    cdef void _subscribe_traded_pairs_feed(self)
    cdef void _subscribe_pair_independent_feed(self)
    cdef void _subscribe_watched_pairs_feed(self)
    cdef bint _should_run_candle_feed(self)
    cdef void _remove_all_feeds(self)
    cdef void _remove_feed(self, object feed)
    cdef void _fix_signal_handler(self)
    cdef void _fix_logger(self)
    cdef list _convert_book_prices_to_orders(self, dict book_prices, str book_side)
    cdef str _parse_order_type(self, str raw_order_type)
    cdef str _parse_order_status(self, str raw_order_status)
    cdef str _parse_order_side(self, str raw_order_side)
    cdef void _register_previous_open_candle(self, object time_frame, str symbol, list candle)
    cdef list _get_previous_open_candle(self, object time_frame, str symbol)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import logging

import ccxt
import octobot_commons.logging as commons_logging

import octobot_trading.exchanges.connectors.abstract_websocket_connector as abstract_websocket_connector


class CCXTWebsocketConnector(abstract_websocket_connector.AbstractWebsocketConnector):
    def __init__(self,
                 config: object,
                 exchange_manager: object,
                 timeout: int = 120,
                 timeout_interval: int = 5):
        super().__init__(config, exchange_manager, timeout, timeout_interval)
        commons_logging.set_logging_level(self.LOGGERS, logging.WARNING)

        self.async_ccxt_client = None
        self.ccxt_client = None
        self.use_testnet = self.exchange_manager.is_sandboxed

    def initialize(self, currencies=None, pairs=None, time_frames=None, channels=None):
        self.async_ccxt_client = self.get_ccxt_async_client()()
        self.ccxt_client = getattr(ccxt, self.get_name())()
        self.ccxt_client.load_markets()
        super().initialize(currencies, pairs, time_frames, channels)

    @classmethod
    def is_handling_spot(cls) -> bool:
        return False

    @classmethod
    def is_handling_margin(cls) -> bool:
        return False

    @classmethod
    def is_handling_future(cls) -> bool:
        return False

    @classmethod
    def get_ccxt_async_client(cls):
        raise NotImplementedError("get_ccxt_async_client is not implemented")

    def get_pair_from_exchange(self, pair):
        try:
            return self.ccxt_client.market(pair)["symbol"]
        except ccxt.errors.BadSymbol:
            try:
                return self.ccxt_client.markets_by_id[pair]["symbol"]
            except KeyError:
                self.logger.error(f"Failed to get market of {pair}")
                return None

    def get_exchange_pair(self, pair):
        if pair in self.ccxt_client.symbols:
            try:
                return self.ccxt_client.market(pair)["id"]
            except KeyError:
                raise KeyError(f'{pair} is not supported on {self.get_name()}')
        else:
            raise ValueError(f'{pair} is not supported on {self.get_name()}')



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library


from octobot_trading.storage import abstract_storage
from octobot_trading.storage.abstract_storage import (
    AbstractStorage,
)


from octobot_trading.storage import trades_storage
from octobot_trading.storage.trades_storage import (
    TradesStorage,
)


from octobot_trading.storage import portfolio_storage
from octobot_trading.storage.portfolio_storage import (
    PortfolioStorage,
)


from octobot_trading.storage import candles_storage
from octobot_trading.storage.candles_storage import (
    CandlesStorage,
)


from octobot_trading.storage import transactions_storage
from octobot_trading.storage.transactions_storage import (
    TransactionsStorage,
)


from octobot_trading.storage import storage_manager
from octobot_trading.storage.storage_manager import (
    StorageManager,
)


__all__ = ["AbstractStorage", "TradesStorage", "PortfolioStorage", "CandlesStorage", "TransactionsStorage", "StorageManager"]



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

cimport octobot_trading.exchanges.exchange_manager as exchange_manager
cimport octobot_trading.util as util


cdef class Trader(util.Initializable):
    cdef dict config

    cdef public object risk
    cdef public bint allow_artificial_orders

    cdef public str trader_type_str

    cdef public bint simulate
    cdef public bint is_enabled

    cdef public object logger

    cdef public exchange_manager.ExchangeManager exchange_manager

    # methods
    cpdef void clear(self)
    cpdef str parse_order_id(self, str order_id)
    cpdef object set_risk(self, object risk)
    cpdef object convert_order_to_trade(self, object order)

    cdef bint _has_open_position(self, str symbol)



# pylint: disable=E0611
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import contextlib
import decimal
import typing
import copy

import ccxt.async_support as ccxt
from octobot_commons import enums as common_enums
from octobot_commons import number_util

import octobot_trading.enums as enums
import octobot_trading.errors as errors
import octobot_trading.exchanges.types as exchanges_types
import octobot_trading.exchanges.util as exchanges_util
import octobot_trading.exchanges.connectors as exchange_connectors
import octobot_trading.personal_data as personal_data
from octobot_trading.enums import ExchangeConstantsOrderColumns as ecoc


#TODO remove
class SpotCCXTExchange(exchanges_types.SpotExchange):
    ORDER_NON_EMPTY_FIELDS = [ecoc.ID.value, ecoc.TIMESTAMP.value, ecoc.SYMBOL.value, ecoc.TYPE.value,
                              ecoc.SIDE.value, ecoc.PRICE.value, ecoc.AMOUNT.value, ecoc.STATUS.value]
    ORDER_REQUIRED_FIELDS = ORDER_NON_EMPTY_FIELDS + [ecoc.REMAINING.value]
    CONNECTOR_CLASS = exchange_connectors.CCXTExchange

    def __init__(self, config, exchange_manager):
        super().__init__(config, exchange_manager)
        self.connector = self.CONNECTOR_CLASS(
            config,
            exchange_manager,
            additional_ccxt_config=self.get_additional_connector_config()
        )

        self.connector.client.options['defaultType'] = self.get_default_type()

    async def initialize_impl(self):
        await self.connector.initialize()
        self.symbols = self.connector.symbols
        self.time_frames = self.connector.time_frames

    async def stop(self) -> None:
        await self.connector.stop()
        self.exchange_manager = None

    @classmethod
    def is_supporting_exchange(cls, exchange_candidate_name) -> bool:    # move to connector
        return cls.CONNECTOR_CLASS.is_supporting_exchange(exchange_candidate_name)

    def get_default_type(self):
        # keep default value
        return self.connector.client.options['defaultType']

    async def create_order(self, order_type: enums.TraderOrderType, symbol: str, quantity: decimal.Decimal,
                           price: decimal.Decimal = None, stop_price: decimal.Decimal = None,
                           side: enums.TradeOrderSide = None, current_price: decimal.Decimal = None,
                           params: dict = None) \
            -> typing.Optional[dict]:
        async with self._order_operation(order_type, symbol, quantity, price, stop_price):
            created_order = await self._create_order_with_retry(order_type, symbol, quantity,
                                                                price, side, current_price, params)
            return await self._verify_order(created_order, order_type, symbol, price)
        return None

    async def edit_order(self, order_id: str, order_type: enums.TraderOrderType, symbol: str,
                         quantity: decimal.Decimal, price: decimal.Decimal,
                         stop_price: decimal.Decimal = None, side: enums.TradeOrderSide = None,
                         current_price: decimal.Decimal = None,
                         params: dict = None):
        # Note: on most exchange, this implementation will just replace the order by cancelling the one
        # which id is given and create a new one
        async with self._order_operation(order_type, symbol, quantity, price, stop_price):
            float_quantity = float(quantity)
            float_price = float(price)
            float_stop_price = None if stop_price is None else float(stop_price)
            float_current_price = None if current_price is None else float(current_price)
            side = None if side is None else side.value
            params = {} if params is None else params
            params.update(self.exchange_manager.exchange_backend.get_orders_parameters(None))
            edited_order = await self._edit_order(order_id, order_type, symbol, quantity=float_quantity,
                                                  price=float_price, stop_price=float_stop_price, side=side,
                                                  current_price=float_current_price, params=params)
            return await self._verify_order(edited_order, order_type, symbol, price)
        return None

    async def _edit_order(self, order_id: str, order_type: enums.TraderOrderType, symbol: str,
                          quantity: float, price: float, stop_price: float = None, side: str = None,
                          current_price: float = None, params: dict = None):
        ccxt_order_type = self.connector.get_ccxt_order_type(order_type)
        price_to_use = price
        if ccxt_order_type == enums.TradeOrderType.MARKET.value:
            # can't set price in market orders
            price_to_use = None
        # do not use keyword arguments here as default ccxt edit order is passing *args (and not **kwargs)
        return await self.connector.client.edit_order(order_id, symbol, ccxt_order_type, side,
                                                      quantity, price_to_use, params)

    @contextlib.asynccontextmanager
    async def _order_operation(self, order_type, symbol, quantity, price, stop_price):
        try:
            yield
        except ccxt.InsufficientFunds as e:
            self.log_order_creation_error(e, order_type, symbol, quantity, price, stop_price)
            self.logger.warning(str(e))
            raise errors.MissingFunds(e)
        except ccxt.NotSupported:
            raise errors.NotSupported
        except Exception as e:
            self.log_order_creation_error(e, order_type, symbol, quantity, price, stop_price)
            self.logger.exception(e, False, f"Unexpected error during order operation: {e}")

    async def _verify_order(self, created_order, order_type, symbol, price, params=None):
        # some exchanges are not returning the full order details on creation: fetch it if necessary
        if created_order and not self._ensure_order_details_completeness(created_order):
            if ecoc.ID.value in created_order:
                params = params or {}
                created_order = await self.exchange_manager.exchange.get_order(created_order[ecoc.ID.value], symbol,
                                                                               params=params)

        # on some exchange, market order are not including price, add it manually to ensure uniformity
        if created_order[ecoc.PRICE.value] is None and price is not None:
            created_order[ecoc.PRICE.value] = float(price)

        return self.clean_order(created_order)

    async def _create_order_with_retry(self, order_type, symbol, quantity: decimal.Decimal,
                                       price: decimal.Decimal, side: enums.TradeOrderSide,
                                       current_price: decimal.Decimal, params) -> dict:
        try:
            return await self._create_specific_order(order_type, symbol, quantity, price=price, side=side,
                                                     current_price=current_price, params=params)
        except (ccxt.InvalidOrder, ccxt.BadRequest) as e:
            # can be raised when exchange precision/limits rules change
            self.logger.debug(f"Failed to create order ({e}) : order_type: {order_type}, symbol: {symbol}. "
                              f"This might be due to an update on {self.name} market rules. Fetching updated rules.")
            await self.connector.client.load_markets(reload=True)
            # retry order creation with updated markets (ccxt will use the updated market values)
            return await self._create_specific_order(order_type, symbol, quantity, price=price, side=side,
                                                     current_price=current_price, params=params)

    def _ensure_order_details_completeness(self, order, order_required_fields=None, order_non_empty_fields=None):
        if order_required_fields is None:
            order_required_fields = self.ORDER_REQUIRED_FIELDS
        if order_non_empty_fields is None:
            order_non_empty_fields = self.ORDER_NON_EMPTY_FIELDS
        # ensure all order_required_fields are present and all order_non_empty_fields are not empty
        return all(key in order for key in order_required_fields) and \
            all(order[key] for key in order_non_empty_fields)

    async def _create_specific_order(self, order_type, symbol, quantity: decimal.Decimal, price: decimal.Decimal = None,
                                     side: enums.TradeOrderSide = None, current_price: decimal.Decimal = None,
                                     params=None) -> dict:
        created_order = None
        float_quantity = float(quantity)
        float_price = float(price)
        float_current_price = float(current_price)
        side = None if side is None else side.value
        params = {} if params is None else params
        params.update(self.exchange_manager.exchange_backend.get_orders_parameters(None))
        if order_type == enums.TraderOrderType.BUY_MARKET:
            created_order = await self._create_market_buy_order(symbol, float_quantity, price=float_price,
                                                                params=params)
        elif order_type == enums.TraderOrderType.BUY_LIMIT:
            created_order = await self._create_limit_buy_order(symbol, float_quantity, price=float_price,
                                                               params=params)
        elif order_type == enums.TraderOrderType.SELL_MARKET:
            created_order = await self._create_market_sell_order(symbol, float_quantity, price=float_price,
                                                                 params=params)
        elif order_type == enums.TraderOrderType.SELL_LIMIT:
            created_order = await self._create_limit_sell_order(symbol, float_quantity, price=float_price,
                                                                params=params)
        elif order_type == enums.TraderOrderType.STOP_LOSS:
            created_order = await self._create_market_stop_loss_order(symbol, float_quantity, price=float_price,
                                                                      side=side, current_price=float_current_price,
                                                                      params=params)
        elif order_type == enums.TraderOrderType.STOP_LOSS_LIMIT:
            created_order = await self._create_limit_stop_loss_order(symbol, float_quantity, price=float_price,
                                                                     side=side, params=params)
        elif order_type == enums.TraderOrderType.TAKE_PROFIT:
            created_order = await self._create_market_take_profit_order(symbol, float_quantity, price=float_price,
                                                                        side=side, params=params)
        elif order_type == enums.TraderOrderType.TAKE_PROFIT_LIMIT:
            created_order = await self._create_limit_take_profit_order(symbol, float_quantity, price=float_price,
                                                                       side=side, params=params)
        elif order_type == enums.TraderOrderType.TRAILING_STOP:
            created_order = await self._create_market_trailing_stop_order(symbol, float_quantity, price=float_price,
                                                                          side=side, params=params)
        elif order_type == enums.TraderOrderType.TRAILING_STOP_LIMIT:
            created_order = await self._create_limit_trailing_stop_order(symbol, float_quantity, price=float_price,
                                                                         side=side, params=params)
        return created_order

    async def _create_market_buy_order(self, symbol, quantity, price=None, params=None) -> dict:
        return await self.connector.client.create_market_buy_order(symbol, quantity, params=params)

    async def _create_limit_buy_order(self, symbol, quantity, price=None, params=None) -> dict:
        return await self.connector.client.create_limit_buy_order(symbol, quantity, price, params=params)

    async def _create_market_sell_order(self, symbol, quantity, price=None, params=None) -> dict:
        return await self.connector.client.create_market_sell_order(symbol, quantity, params=params)

    async def _create_limit_sell_order(self, symbol, quantity, price=None, params=None) -> dict:
        return await self.connector.client.create_limit_sell_order(symbol, quantity, price, params=params)

    async def _create_market_stop_loss_order(self, symbol, quantity, price, side, current_price, params=None) -> dict:
        raise NotImplementedError("_create_market_stop_loss_order is not implemented")

    async def _create_limit_stop_loss_order(self, symbol, quantity, price=None, side=None, params=None) -> dict:
        raise NotImplementedError("_create_limit_stop_loss_order is not implemented")

    async def _create_market_take_profit_order(self, symbol, quantity, price=None, side=None, params=None) -> dict:
        raise NotImplementedError("_create_market_take_profit_order is not implemented")

    async def _create_limit_take_profit_order(self, symbol, quantity, price=None, side=None, params=None) -> dict:
        raise NotImplementedError("_create_limit_take_profit_order is not implemented")

    async def _create_market_trailing_stop_order(self, symbol, quantity, price=None, side=None, params=None) -> dict:
        raise NotImplementedError("_create_market_trailing_stop_order is not implemented")

    async def _create_limit_trailing_stop_order(self, symbol, quantity, price=None, side=None, params=None) -> dict:
        raise NotImplementedError("_create_limit_trailing_stop_order is not implemented")

    def get_exchange_current_time(self):
        return self.connector.get_exchange_current_time()

    def get_uniform_timestamp(self, timestamp):
        return self.connector.get_uniform_timestamp(timestamp)

    def get_market_status(self, symbol, price_example=None, with_fixer=True):
        """
        Override using get_fixed_market_status in exchange tentacle if the default market status is not as expected
        """
        return self.connector.get_market_status(symbol, price_example=price_example, with_fixer=with_fixer)

    def get_fixed_market_status(self, symbol, price_example=None, with_fixer=True, remove_price_limits=False):
        """
        Use this method in local get_market_status overrides when market status has to be fixed by
        calling _fix_market_status.
        Changes PRECISION_AMOUNT and PRECISION_PRICE from decimals to integers
        (use number of digits instead of price example) by default.
        Override _fix_market_status to change other elements
        """
        market_status = self._fix_market_status(
            copy.deepcopy(
                self.connector.get_market_status(symbol, with_fixer=False)
            ),
            remove_price_limits=remove_price_limits
        )
        if with_fixer:
            return exchanges_util.ExchangeMarketStatusFixer(market_status, price_example).market_status
        return market_status

    def _fix_market_status(self, market_status, remove_price_limits=False):
        """
        Overrite if necessary
        """
        market_status[enums.ExchangeConstantsMarketStatusColumns.PRECISION.value][
            enums.ExchangeConstantsMarketStatusColumns.PRECISION_AMOUNT.value] = number_util.get_digits_count(
            market_status[enums.ExchangeConstantsMarketStatusColumns.PRECISION.value][
                enums.ExchangeConstantsMarketStatusColumns.PRECISION_AMOUNT.value]
        )
        market_status[enums.ExchangeConstantsMarketStatusColumns.PRECISION.value][
            enums.ExchangeConstantsMarketStatusColumns.PRECISION_PRICE.value] = number_util.get_digits_count(
            market_status[enums.ExchangeConstantsMarketStatusColumns.PRECISION.value][
                enums.ExchangeConstantsMarketStatusColumns.PRECISION_PRICE.value]
        )
        if remove_price_limits:
            market_status[enums.ExchangeConstantsMarketStatusColumns.LIMITS.value][
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_PRICE.value][
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_PRICE_MIN.value] = None
            market_status[enums.ExchangeConstantsMarketStatusColumns.LIMITS.value][
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_PRICE.value][
                enums.ExchangeConstantsMarketStatusColumns.LIMITS_PRICE_MAX.value] = None

        return market_status

    async def get_balance(self, **kwargs: dict):
        return await self.connector.get_balance(**kwargs)

    async def get_symbol_prices(self, symbol: str, time_frame: common_enums.TimeFrames, limit: int = None,
                                **kwargs: dict) -> typing.Optional[list]:
        return await self.connector.get_symbol_prices(symbol=symbol, time_frame=time_frame, limit=limit, **kwargs)

    async def get_kline_price(self, symbol: str, time_frame: common_enums.TimeFrames, **kwargs: dict) -> typing.Optional[list]:
        return await self.connector.get_kline_price(symbol=symbol, time_frame=time_frame, **kwargs)

    async def get_order_book(self, symbol: str, limit: int = 5, **kwargs: dict) -> typing.Optional[dict]:
        return await self.connector.get_order_book(symbol=symbol, limit=limit, **kwargs)

    async def get_recent_trades(self, symbol: str, limit: int = 50, **kwargs: dict) -> typing.Optional[list]:
        return await self.connector.get_recent_trades(symbol=symbol, limit=limit, **kwargs)

    async def get_price_ticker(self, symbol: str, **kwargs: dict) -> typing.Optional[dict]:
        return await self.connector.get_price_ticker(symbol=symbol, **kwargs)

    async def get_all_currencies_price_ticker(self, **kwargs: dict) -> typing.Optional[list]:
        return await self.connector.get_all_currencies_price_ticker(**kwargs)

    async def get_order(self, order_id: str, symbol: str = None, **kwargs: dict) -> dict:
        return await self.connector.get_order(symbol=symbol, order_id=order_id, **kwargs)

    async def get_order_from_open_and_closed_orders(self, order_id: str, symbol: str = None, **kwargs: dict) -> dict:
        for order in await self.get_open_orders(symbol, **kwargs):
            if order[enums.ExchangeConstantsOrderColumns.ID.value] == order_id:
                return order
        for order in await self.get_closed_orders(symbol, **kwargs):
            if order[enums.ExchangeConstantsOrderColumns.ID.value] == order_id:
                return order
        return None  # OrderNotFound

    async def get_order_from_trades(self, symbol, order_id, order_to_update=None):
        order_to_update = order_to_update or {}
        trades = await self.get_my_recent_trades(symbol)
        # usually the right trade is within the last ones
        for trade in trades[::-1]:
            if trade[ecoc.ORDER.value] == order_id:
                order_to_update[ecoc.INFO.value] = trade[ecoc.INFO.value]
                order_to_update[ecoc.ID.value] = order_id
                order_to_update[ecoc.SYMBOL.value] = symbol
                order_to_update[ecoc.TYPE.value] = trade[ecoc.TYPE.value]
                order_to_update[ecoc.AMOUNT.value] = trade[ecoc.AMOUNT.value]
                order_to_update[ecoc.DATETIME.value] = trade[ecoc.DATETIME.value]
                order_to_update[ecoc.SIDE.value] = trade[ecoc.SIDE.value]
                order_to_update[ecoc.TAKERORMAKER.value] = trade[ecoc.TAKERORMAKER.value]
                order_to_update[ecoc.PRICE.value] = trade[ecoc.PRICE.value]
                order_to_update[ecoc.TIMESTAMP.value] = order_to_update.get(ecoc.TIMESTAMP.value,
                                                                            trade[ecoc.TIMESTAMP.value])
                order_to_update[ecoc.STATUS.value] = enums.OrderStatus.FILLED.value
                order_to_update[ecoc.FILLED.value] = trade[ecoc.AMOUNT.value]
                order_to_update[ecoc.COST.value] = trade[ecoc.COST.value]
                order_to_update[ecoc.REMAINING.value] = 0
                order_to_update[ecoc.FEE.value] = trade[ecoc.FEE.value]
                return order_to_update
        return None  #OrderNotFound

    async def get_all_orders(self, symbol: str = None, since: int = None, limit: int = None, **kwargs: dict) -> list:
        return await self.connector.get_all_orders(symbol=symbol, since=since, limit=limit, **kwargs)

    async def get_open_orders(self, symbol: str = None, since: int = None, limit: int = None, **kwargs: dict) -> list:
        return await self.connector.get_open_orders(symbol=symbol, since=since, limit=limit, **kwargs)

    async def get_closed_orders(self, symbol: str = None, since: int = None, limit: int = None, **kwargs: dict) -> list:
        return await self.connector.get_closed_orders(symbol=symbol, since=since, limit=limit, **kwargs)

    async def get_my_recent_trades(self, symbol: str = None, since: int = None, limit: int = None, **kwargs: dict) -> list:
        return await self.connector.get_my_recent_trades(symbol=symbol, since=since, limit=limit, **kwargs)

    async def cancel_order(self, order_id: str, symbol: str = None, **kwargs: dict) -> bool:
        return await self.connector.cancel_order(symbol=symbol, order_id=order_id, **kwargs)

    def get_trade_fee(self, symbol, order_type, quantity, price, taker_or_maker):
        return self.connector.get_trade_fee(symbol, order_type, quantity, price, taker_or_maker)

    def get_fees(self, symbol):
        return self.connector.get_fees(symbol)

    def get_pair_from_exchange(self, pair) -> str:
        return self.connector.get_pair_from_exchange(pair)

    def get_split_pair_from_exchange(self, pair) -> (str, str):
        return self.connector.get_split_pair_from_exchange(pair)

    def get_exchange_pair(self, pair) -> str:
        return self.connector.get_exchange_pair(pair)

    def get_pair_cryptocurrency(self, pair) -> str:
        return self.connector.get_pair_cryptocurrency(pair)

    def get_default_balance(self):
        return self.connector.get_default_balance()

    def get_rate_limit(self):
        return self.connector.get_rate_limit()

    async def switch_to_account(self, account_type: enums.AccountTypes):
        return await self.connector.switch_to_account(account_type=account_type)

    def parse_balance(self, balance):
        return personal_data.parse_decimal_portfolio(self.connector.parse_balance(balance))

    def parse_trade(self, trade):
        return self.connector.parse_trade(trade)

    def parse_order(self, order):
        return self.connector.parse_order(order)

    def parse_ticker(self, ticker):
        return self.connector.parse_ticker(ticker)

    def parse_ohlcv(self, ohlcv):
        return self.connector.parse_ohlcv(ohlcv)

    def parse_order_book(self, order_book):
        return self.connector.parse_order_book(order_book)

    def parse_order_book_ticker(self, order_book_ticker):
        return self.connector.parse_order_book_ticker(order_book_ticker)

    def parse_timestamp(self, data_dict, timestamp_key, default_value=None, ms=False):
        return self.connector.parse_timestamp(data_dict, timestamp_key, default_value=default_value, ms=ms)

    def parse_currency(self, currency):
        return self.connector.parse_currency(currency)

    def parse_order_id(self, order):
        return self.connector.parse_order_id(order)

    def parse_order_symbol(self, order):
        return self.connector.parse_order_symbol(order)

    def parse_status(self, status):
        return self.connector.parse_status(status)

    def parse_side(self, side):
        return self.connector.parse_side(side)

    def parse_account(self, account):
        return self.connector.parse_account(account)

    def clean_recent_trade(self, recent_trade):
        return self.connector.clean_recent_trade(recent_trade)

    def clean_trade(self, trade):
        return self.connector.clean_trade(trade)

    def clean_order(self, order):
        return self.connector.clean_order(order)



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

cpdef bint is_valid(object element, object key)
cpdef object get_min_max_amounts(dict symbol_market, object default_value=*)
cpdef bint check_cost(double total_order_price, object min_cost)
cpdef object total_fees_from_order_dict(dict order_dict, str currency)
cpdef object get_fees_for_currency(object fee, str currency)
cpdef dict parse_raw_fees(object raw_fees)
cpdef object parse_order_status(dict raw_order)
cpdef object parse_is_cancelled(dict raw_order)
cpdef object get_max_order_quantity_for_price(object position, object available_quantity,
                                              object price, object side, str symbol)
cpdef object get_pnl_transaction_source_from_order(object order)
cpdef bint is_stop_order(object order_type)
cpdef bint is_associated_pending_order(object pending_order, object created_order)
cpdef object get_order_quantity_currency(object exchange_manager, str symbol, object side)
cpdef str generate_order_id()



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport octobot_trading.exchanges as exchanges

cdef class ExchangeBuilder:
    cdef object logger
    cdef public exchanges.ExchangeManager exchange_manager

    cdef dict config

    cdef bint _is_using_trading_modes

    cdef public str exchange_name
    cdef str _matrix_id

    """
    Builder methods
    """
    cpdef ExchangeBuilder is_backtesting(self, object backtesting_instance)
    cpdef ExchangeBuilder is_sandboxed(self, bint sandboxed)
    cpdef ExchangeBuilder is_simulated(self)
    cpdef ExchangeBuilder is_loading_markets(self, bint is_loading_markets)
    cpdef ExchangeBuilder is_real(self)
    cpdef ExchangeBuilder is_using_exchange_type(self, str exchange_type)
    cpdef ExchangeBuilder enable_storage(self, bint enabled)
    cpdef ExchangeBuilder is_margin(self, bint use_margin=*)
    cpdef ExchangeBuilder is_exchange_only(self)
    cpdef ExchangeBuilder is_ignoring_config(self)
    cpdef ExchangeBuilder is_without_auth(self)
    cpdef ExchangeBuilder is_checking_credentials(self, bint check_credentials)
    cpdef ExchangeBuilder use_tentacles_setup_config(self, object tentacles_setup_config)
    cpdef ExchangeBuilder set_bot_id(self, str bot_id)
    cpdef ExchangeBuilder disable_trading_mode(self)
    cpdef ExchangeBuilder has_matrix(self, str matrix_id)

    cdef void _register_trading_modes_requirements(self, object trading_mode_class, object tentacles_setup_config)
    cdef object _ensure_exchange_compatibility(self)
    cdef object _ensure_trading_mode_compatibility(self, object trading_mode_class)

cpdef ExchangeBuilder create_exchange_builder_instance(object config, str exchange_name)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import octobot_trading.personal_data.portfolios.asset as asset
import octobot_trading.constants as constants
import octobot_trading.enums as enums
import octobot_commons.constants as commons_constants
import octobot_commons.logging as commons_logging
import octobot_commons.symbols as symbol_util
import numpy as numpy


def parse_decimal_portfolio(portfolio):
    decimal_portfolio = {}
    for symbol, symbol_balance in portfolio.items():
        if isinstance(symbol_balance, dict):
            decimal_portfolio[symbol] = {}
            portfolio_to_fill = decimal_portfolio[symbol]
            for balance_type, balance_val in symbol_balance.items():
                if isinstance(balance_val, (int, float, decimal.Decimal)):
                    portfolio_to_fill[balance_type] = decimal.Decimal(str(balance_val))
    return decimal_portfolio


def parse_decimal_config_portfolio(portfolio):
    return {
        symbol: decimal.Decimal(str(symbol_balance))
        for symbol, symbol_balance in portfolio.items()
    }


def portfolio_to_float(portfolio):
    float_portfolio = {}
    for symbol, symbol_balance in portfolio.items():
        if isinstance(symbol_balance, asset.Asset):
            float_portfolio[symbol] = {
                commons_constants.PORTFOLIO_AVAILABLE: float(symbol_balance.available),
                commons_constants.PORTFOLIO_TOTAL: float(symbol_balance.total)
            }
    return float_portfolio


def get_draw_down(exchange_manager):
    """
    Draw down is the lowest portfolio value in % ever reached during a run
    :return: the Draw down value
    """
    draw_down = constants.ZERO
    if exchange_manager.is_future:
        try:
            value_currency = exchange_manager.exchange_personal_data.portfolio_manager.reference_market
            draw_down_pair = exchange_manager.exchange_config.traded_symbol_pairs[0]
            if exchange_manager.exchange_personal_data.positions_manager.get_symbol_position(
                draw_down_pair,
                enums.PositionSide.BOTH
            ).symbol_contract.is_inverse_contract():
                value_currency = symbol_util.parse_symbol(draw_down_pair).base
            if exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder.origin_portfolio \
                    is None:
                return constants.ZERO
            origin_value = exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder \
                .origin_portfolio.portfolio[value_currency].total
            portfolio_history = [origin_value]
            for transaction in exchange_manager.exchange_personal_data.transactions_manager.transactions.values():
                current_pnl = transaction.quantity if hasattr(transaction, "quantity") else transaction.realised_pnl
                portfolio_history.append(portfolio_history[-1] + current_pnl)

                current_draw_down = constants.ONE_HUNDRED - \
                    (portfolio_history[-1] / (max(portfolio_history) / constants.ONE_HUNDRED))

                draw_down = current_draw_down if current_draw_down > draw_down else draw_down
        except Exception as e:
            commons_logging.get_logger(__name__).exception(e, True, f"Error when computing draw down: {e}")
    return draw_down


async def get_coefficient_of_determination_data(transactions, start_balance,
                                                use_high_instead_of_end_balance=True,
                                                x_as_trade_count=True):
    if transactions:
        # get realized portfolio history

        pnl_history = [start_balance]
        pnl_history_times = []
        for transaction in transactions:
            current_pnl = None
            if hasattr(transaction, "quantity"):
                current_pnl = float(transaction.quantity)
                pnl_history_times.append(transaction.creation_time)
            elif hasattr(transaction, 'realised_pnl'):
                current_pnl = float(transaction.realised_pnl)
                pnl_history_times.append(transaction.creation_time)
            elif isinstance(transaction, dict):
                if transaction["quantity"]:
                    current_pnl = transaction["quantity"]
                    pnl_history_times.append(transaction["x"])
                elif transaction['realised_pnl']:
                    current_pnl = transaction['realised_pnl']
                    pnl_history_times.append(transaction["x"])

            if current_pnl is not None:
                pnl_history.append(pnl_history[-1] + current_pnl)

        if pnl_history_times:

            pnl_history_for_every_candle = None
            # either use trade to trade basis or candle to candle basis
            if x_as_trade_count is True:
                start_time = pnl_history_times[0]
                pnl_history_times.insert(0, start_time)
                end_time = pnl_history_times[-1]
                data_length = len(pnl_history)

            else:  # calculate pnl history for every candle
                raise NotImplementedError("x_as_trade_count=False is not implemented")
            # calculate best case data (exponential growth)
            end_balance = pnl_history[-1]

            if start_balance > end_balance:
                # if we end up with a negative balance we can't compute this
                return None, None, None, None, None

            if use_high_instead_of_end_balance:
                end_value = max(pnl_history)
            else:  # use the end balance
                end_value = end_balance

            def get_ideal_value(linear_growth, adj1, adj2):
                return ((linear_growth + adj1) ** pw) * adj2

            x = [start_time, end_time]
            y = [start_balance, end_value]

            pw = 15
            A = numpy.exp(numpy.log(y[0] / y[1]) / pw)
            a = (x[0] - x[1] * A) / (A - 1)
            b = y[0] / (x[0] + a) ** pw

            linear_growth = numpy.linspace(start_time, end_time, data_length)
            best_case = get_ideal_value(linear_growth, a, b)
            pnl_data = pnl_history_for_every_candle or pnl_history
            return list(best_case), pnl_data, start_balance, end_balance, pnl_history_times
    return None, None, None, None, None


async def get_coefficient_of_determination(exchange_manager, use_high_instead_of_end_balance=True):
    """
    Calculates proximity to the best case growth for the current run (best growth being associated to an exponential
    curve). The closer the actual result, the higher the coefficient_of_determination (R squared) will be.
    Return 0 if we end up with less money that we had to begin with
    :param use_high_instead_of_end_balance: best case exponential growth based on end balance or highest balance
    """
    coefficient_of_determination = 0
    # get data the data necessary to compute the coefficient_of_determination
    origin_portfolio = portfolio_to_float(
        exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder.origin_portfolio.portfolio
    )
    start_balance = origin_portfolio[
        exchange_manager.exchange_personal_data.portfolio_manager.reference_market][commons_constants.PORTFOLIO_TOTAL]

    best_case, pnl_data, start_balance, _, _ \
        = await get_coefficient_of_determination_data(transactions=exchange_manager.exchange_personal_data.
                                                      transactions_manager.transactions.values(),
                                                      start_balance=start_balance,
                                                      use_high_instead_of_end_balance=use_high_instead_of_end_balance)

    if pnl_data:
        # calculate r²
        corr_matrix = numpy.corrcoef(best_case, pnl_data)
        corr = corr_matrix[0, 1]
        coefficient_of_determination = corr ** 2

    return round(coefficient_of_determination, 3)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from octobot_trading.exchanges cimport exchange_channels
from octobot_trading.exchanges.exchange_channels cimport (
    requires_refresh_trigger,
)
from octobot_trading.exchanges.exchanges cimport (
    ExchangeConfiguration,
    Exchanges,
)

from octobot_trading.exchanges.config.exchange_config_data cimport (
    ExchangeConfig,
)

from octobot_trading.exchanges.traders cimport (
    Trader,
    TraderSimulator,
)

from octobot_trading.exchanges.abstract_exchange cimport (
    AbstractExchange,
)

from octobot_trading.exchanges.abstract_websocket_exchange cimport (
    AbstractWebsocketExchange,
)

from octobot_trading.exchanges.basic_exchange_wrapper cimport (
    BasicExchangeWrapper,
)

from octobot_trading.exchanges.exchange_manager cimport (
    ExchangeManager,
)
from octobot_trading.exchanges.util.exchange_util cimport (
    get_margin_exchange_class,
    get_future_exchange_class,
    get_spot_exchange_class,
    get_order_side,
)
from octobot_trading.exchanges.util cimport (
    ExchangeMarketStatusFixer,
    is_ms_valid,
    force_disable_web_socket,
    check_web_socket_config,
    search_websocket_class,
)
from octobot_trading.exchanges.types cimport (
    FutureExchange,
    WebSocketExchange,
    MarginExchange,
    SpotExchange,
)
from octobot_trading.exchanges.util cimport (
    ExchangeMarketStatusFixer,
    is_ms_valid,
    get_margin_exchange_class,
    get_future_exchange_class,
    get_spot_exchange_class,
    get_order_side,
    force_disable_web_socket,
    check_web_socket_config,
    search_websocket_class,
    get_exchange_type,
    get_default_exchange_type,
    get_supported_exchange_types,
    get_exchange_class_from_name,
)

from octobot_trading.exchanges cimport implementations
from octobot_trading.exchanges.implementations cimport (
    SpotExchangeSimulator,
    SpotCCXTExchange,
    FutureExchangeSimulator,
    FutureCCXTExchange,
    MarginExchangeSimulator,
    MarginCCXTExchange,
    CCXTWebSocketExchange,
    CryptofeedWebSocketExchange,
)

from octobot_trading.exchanges.exchange_builder cimport (
    ExchangeBuilder,
)

from octobot_trading.exchanges cimport connectors
from octobot_trading.exchanges.connectors cimport (
    ExchangeSimulator,
    CCXTExchange,
    CCXTWebsocketConnector,
    AbstractWebsocketConnector,
    CryptofeedWebsocketConnector,
)

from octobot_trading.exchanges cimport abstract_websocket_exchange
from octobot_trading.exchanges.abstract_websocket_exchange cimport (
    AbstractWebsocketExchange,
)

__all__ = [
    "requires_refresh_trigger",
    "ExchangeConfig",
    "ExchangeManager",
    "ExchangeBuilder",
    "ExchangeConfiguration",
    "Exchanges",
    "get_margin_exchange_class",
    "get_future_exchange_class",
    "get_spot_exchange_class",
    "get_order_side",
    "AbstractExchange",
    "TraderSimulator",
    "Trader",
    "ExchangeSimulator",
    "CCXTExchange",
    "AbstractWebsocketConnector",
    "CCXTWebsocketConnector",
    "AbstractWebsocketExchange",
    "BasicExchangeWrapper",
    "FutureExchange",
    "MarginExchange",
    "SpotExchange",
    "WebSocketExchange",
    "ExchangeMarketStatusFixer",
    "is_ms_valid",
    "AbstractWebsocketExchange",
    "force_disable_web_socket",
    "check_web_socket_config",
    "search_websocket_class",
    "get_exchange_type",
    "get_default_exchange_type",
    "get_supported_exchange_types",
    "get_exchange_class_from_name",
    "SpotExchangeSimulator",
    "SpotCCXTExchange",
    "FutureExchangeSimulator",
    "FutureCCXTExchange",
    "MarginExchangeSimulator",
    "MarginCCXTExchange",
    "CCXTWebSocketExchange",
    "CryptofeedWebSocketExchange",
    "CryptofeedWebsocketConnector",
]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import collections

import octobot_commons.logging as logging
import octobot_commons.tree as commons_tree
import octobot_commons.enums as commons_enums

import octobot_trading.enums as enums
import octobot_trading.personal_data as personal_data
import octobot_trading.util as util


class TradesManager(util.Initializable):
    # memory usage for 100000 trades: approx 180 Mo
    MAX_TRADES_COUNT = 100000

    def __init__(self, trader):
        super().__init__()
        self.logger = logging.get_logger(self.__class__.__name__)
        self.trader = trader
        self.trades_initialized = False
        self.trades = collections.OrderedDict()

    async def initialize_impl(self):
        self._reset_trades()
        self.trades_initialized = True
        if self.trader.simulate:
            # force init as there is no trade updater simulator
            for symbol in self.trader.exchange_manager.exchange_config.traded_symbol_pairs:
                self._set_initialized_event(symbol)

    def upsert_trade(self, trade_id, raw_trade):
        if trade_id not in self.trades:
            created_trade = personal_data.create_trade_instance_from_raw(self.trader, raw_trade)
            if created_trade:
                if trade_id in self.trades:
                    self.logger.debug(f"Replacement of an existing trade: {self.trades[trade_id].to_dict()} "
                                      f"by {created_trade.to_dict()} on id: {trade_id}")
                self.trades[trade_id] = created_trade
                self._check_trades_size()
                return True
        return False

    def upsert_trade_instance(self, trade):
        if trade.trade_id not in self.trades:
            self.trades[trade.trade_id] = trade
            self._check_trades_size()

    def has_closing_trade_with_order_id(self, order_id) -> bool:
        for trade in self.trades.values():
            if trade.origin_order_id == order_id and trade.is_closing_order:
                return True
        return False

    def get_total_paid_fees(self):
        total_fees = {}
        for trade in self.trades.values():
            if trade.fee is not None:
                fee_cost = trade.fee[enums.FeePropertyColumns.COST.value]
                fee_currency = trade.fee[enums.FeePropertyColumns.CURRENCY.value]
                if fee_currency in total_fees:
                    total_fees[fee_currency] += fee_cost
                else:
                    total_fees[fee_currency] = fee_cost
            elif trade.status is not enums.OrderStatus.CANCELED:
                self.logger.warning(f"Trade without any registered fee: {trade.to_dict()}")
        return total_fees

    def get_trade(self, trade_id):
        return self.trades[trade_id]

    # private
    def _check_trades_size(self):
        if len(self.trades) > self.MAX_TRADES_COUNT:
            self._remove_oldest_trades(int(self.MAX_TRADES_COUNT / 10))

    def _reset_trades(self):
        self.trades_initialized = False
        self.trades = collections.OrderedDict()

    def _remove_oldest_trades(self, nb_to_remove):
        for _ in range(nb_to_remove):
            self.trades.popitem(last=False)

    def _set_initialized_event(self, symbol):
        # set init in updater as it's the only place we know if we fetched trades or not regardless of trades existence
        commons_tree.EventProvider.instance().trigger_event(
            self.trader.exchange_manager.bot_id, commons_tree.get_exchange_path(
                self.trader.exchange_manager.exchange_name,
                commons_enums.InitializationEventExchangeTopics.TRADES.value
            )
        )

    def clear(self):
        for trade in self.trades.values():
            trade.trader = None
            trade.exchange_manager = None
        self._reset_trades()



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport octobot_backtesting.backtesting as backtesting
cimport octobot_trading.exchanges.abstract_exchange as abstract_exchange

cdef class ExchangeSimulator(abstract_exchange.AbstractExchange):
    cdef public backtesting.Backtesting backtesting

    cdef public list exchange_importers

    cdef public dict current_future_candles

    cdef public bint is_authenticated

    cpdef str get_pair_cryptocurrency(self, str pair)
    cpdef list get_available_time_frames(self)
    cpdef list get_backtesting_data_files(self)
    cpdef list get_time_frames(self, object importer)

    cdef void _read_fees_from_config(self, dict result_fees)

# Should be cythonized with cython 3.0
# cpdef set handles_real_data_for_updater(str channel_type, list available_data_types)

# Should be cythonized with cython 3.0
# cdef bint _are_required_data_available(str channel_type, list available_data_types)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import contextlib

import octobot_commons.logging as logging

import octobot_trading.enums as enums
import octobot_trading.util as util


class State(util.Initializable):
    def __init__(self, is_from_exchange_data):
        super().__init__()

        # default state
        self.state = enums.States.UNKNOWN

        # if this state has been created from exchange data or OctoBot internal mechanism
        self.is_from_exchange_data = is_from_exchange_data

        # state lock
        self.lock = asyncio.Lock()

    def is_pending(self) -> bool:
        """
        :return: True if the state is pending for update
        """
        return self.state is enums.States.UNKNOWN

    def is_refreshing(self) -> bool:
        """
        :return: True if the state is updating
        """
        return self.state is enums.States.REFRESHING

    def is_open(self) -> bool:
        """
        :return: True if the instance is considered as open
        """
        return not self.is_closed()

    def is_closed(self) -> bool:
        """
        :return: True if the instance is considered as closed
        """
        return False

    def get_logger(self):
        """
        :return: the instance logger
        """
        return logging.get_logger(self.__class__.__name__)

    def log_event_message(self, state_message, error=None):
        """
        Log a state event
        """
        self.get_logger().info(state_message.value)

    async def initialize_impl(self) -> None:
        """
        Default State initialization process
        """
        await self.update()

    async def should_be_updated(self) -> bool:
        """
        Defines if the instance should be updated
        :return: True if the instance should be updated when necessary
        """
        return True

    async def update(self) -> None:
        """
        Update the instance state if necessary
        Necessary when the state is not already synchronizing and when the instance should be updated
        Try to fix the pending state or terminate
        """
        if not self.is_refreshing():
            if self.is_pending() and not await self.should_be_updated():
                self.log_event_message(enums.StatesMessages.SYNCHRONIZING)
                await self.synchronize()
            else:
                async with self.lock:
                    await self.terminate()
        else:
            self.log_event_message(enums.StatesMessages.ALREADY_SYNCHRONIZING)

    async def synchronize(self, force_synchronization=False, catch_exception=False) -> None:
        """
        Implement the exchange synchronization process
        Should begin by setting the state to REFRESHING
        Should end by :
        - calling terminate if the state is terminated
        - restoring the initial state if nothing has been changed with synchronization or if sync failed
        :param force_synchronization: When True, for the update of the order from the exchange
        :param catch_exception: When False raises the Exception during synchronize_order instead of catching it silently
        """
        try:
            await self.synchronize_with_exchange(force_synchronization=force_synchronization)
        except Exception as e:
            if catch_exception:
                self.log_event_message(enums.StatesMessages.SYNCHRONIZING_ERROR, error=e)
            else:
                raise

    async def synchronize_with_exchange(self, force_synchronization: bool = False) -> None:
        """
        Ask the exchange to update the order only if the state is not already refreshing
        When the refreshing process starts set the state to enums.States.REFRESHING
        Restore the previous state if the refresh process fails
        :param force_synchronization: When True, for the update of the order from the exchange
        """
        if self.is_refreshing():
            self.log_event_message(enums.StatesMessages.ALREADY_SYNCHRONIZING)
        else:
            async with self.refresh_operation():
                await self._synchronize_with_exchange(force_synchronization=force_synchronization)

    @contextlib.asynccontextmanager
    async def refresh_operation(self):
        self.get_logger().debug("Starting refresh_operation")
        previous_state = self.state
        async with self.lock:
            self.state = enums.States.REFRESHING
        try:
            yield
        finally:
            async with self.lock:
                if self.state is enums.States.REFRESHING:
                    self.state = previous_state
            self.get_logger().debug("Completed refresh_operation")

    async def _synchronize_with_exchange(self, force_synchronization: bool = False) -> None:
        """
        Called when state should be refreshed
        :param force_synchronization: When True, for the update of the order from the exchange
        """
        raise NotImplementedError("_synchronize_with_exchange not implemented")

    async def terminate(self) -> None:
        """
        Implement the state ending process
        Can be portfolio updates, fees request, orders group updates, Trade creation etc...
        """
        raise NotImplementedError("terminate not implemented")

    async def on_refresh_successful(self):
        """
        Called when synchronize succeed to update the instance
        """
        raise NotImplementedError("on_refresh_successful not implemented")

    def clear(self):
        """
        Clear references
        """
        raise NotImplementedError("clear not implemented")



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import octobot_trading.enums
import octobot_trading.personal_data as personal_data


def get_trade_history(exchange_manager, symbol=None, since=None, as_dict=False, include_cancelled=False) -> list:
    return [trade.to_dict() if as_dict else trade
            for trade in exchange_manager.exchange_personal_data.trades_manager.trades.values()
            if _trade_filter(trade, symbol, since, include_cancelled)]


def _trade_filter(trade, symbol=None, timestamp=None, include_cancelled=False) -> bool:
    if trade.status is octobot_trading.enums.OrderStatus.CANCELED and not include_cancelled:
        return False
    if symbol is None and timestamp is None:
        return True
    elif symbol is None and timestamp is not None:
        return _is_trade_after(trade, timestamp)
    elif symbol is not None and timestamp is None:
        return trade.symbol == symbol
    else:
        return trade.symbol == symbol and _is_trade_after(trade, timestamp)


def _is_trade_after(trade, timestamp) -> bool:
    return trade.executed_time > timestamp or trade.canceled_time > timestamp


def get_total_paid_trading_fees(exchange_manager) -> dict:
    return exchange_manager.exchange_personal_data.trades_manager.get_total_paid_fees()


def get_trade_exchange_name(trade) -> str:
    return trade.exchange_manager.get_exchange_name()


def parse_trade_type(dict_trade) -> octobot_trading.enums.TraderOrderType:
    # can use parse_order_type since format is compatible
    return personal_data.parse_order_type(dict_trade)[1]


def trade_to_dict(trade) -> dict:
    return trade.to_dict()


def get_win_rate(exchange_manager) -> decimal.Decimal:
    return personal_data.compute_win_rate(exchange_manager)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_trading.exchange_data cimport funding
from octobot_trading.exchange_data.funding cimport (
    FundingUpdaterSimulator,
    FundingUpdater,
    FundingManager,
    FundingProducer,
    FundingChannel,
)

from octobot_trading.exchange_data cimport kline
from octobot_trading.exchange_data.kline cimport (
    KlineUpdaterSimulator,
    KlineProducer,
    KlineChannel,
    KlineManager,
    KlineUpdater,
)

from octobot_trading.exchange_data cimport ohlcv
from octobot_trading.exchange_data.ohlcv cimport (
    CandlesManager,
    get_symbol_close_candles,
    get_symbol_open_candles,
    get_symbol_high_candles,
    get_symbol_low_candles,
    get_symbol_volume_candles,
    get_symbol_time_candles,
    OHLCVUpdaterSimulator,
    OHLCVProducer,
    OHLCVChannel,
    OHLCVUpdater,
)

from octobot_trading.exchange_data cimport order_book
from octobot_trading.exchange_data.order_book cimport (
    OrderBookUpdater,
    OrderBookProducer,
    OrderBookChannel,
    OrderBookTickerProducer,
    OrderBookTickerChannel,
    OrderBookManager,
    OrderBookUpdaterSimulator,
)

from octobot_trading.exchange_data cimport prices
from octobot_trading.exchange_data.prices cimport (
    MarkPriceUpdaterSimulator,
    MarkPriceProducer,
    MarkPriceChannel,
    PricesManager,
    calculate_mark_price_from_recent_trade_prices,
    MarkPriceUpdater,
    PriceEventsManager,
)

from octobot_trading.exchange_data cimport recent_trades
from octobot_trading.exchange_data.recent_trades cimport (
    RecentTradeProducer,
    RecentTradeChannel,
    LiquidationsProducer,
    LiquidationsChannel,
    RecentTradesManager,
    RecentTradeUpdater,
)

from octobot_trading.exchange_data cimport ticker
from octobot_trading.exchange_data.ticker cimport (
    TickerManager,
    TickerUpdater,
    TickerProducer,
    TickerChannel,
    MiniTickerProducer,
    MiniTickerChannel,
    TickerUpdaterSimulator,
)

from octobot_trading.exchange_data cimport contracts
from octobot_trading.exchange_data.contracts cimport (
    MarginContract,
    FutureContract,
)

from octobot_trading.exchange_data cimport exchange_symbol_data
from octobot_trading.exchange_data.exchange_symbol_data cimport (
    ExchangeSymbolData,
)
from octobot_trading.exchange_data cimport exchange_symbols_data
from octobot_trading.exchange_data.exchange_symbols_data cimport (
    ExchangeSymbolsData,
)

__all__ = [
    "FundingUpdaterSimulator",
    "FundingUpdater",
    "FundingManager",
    "FundingProducer",
    "FundingChannel",
    "KlineUpdaterSimulator",
    "KlineProducer",
    "KlineChannel",
    "KlineManager",
    "KlineUpdater",
    "CandlesManager",
    "get_symbol_close_candles",
    "get_symbol_open_candles",
    "get_symbol_high_candles",
    "get_symbol_low_candles",
    "get_symbol_volume_candles",
    "get_symbol_time_candles",
    "OHLCVUpdaterSimulator",
    "OHLCVProducer",
    "OHLCVChannel",
    "OHLCVUpdater",
    "OrderBookUpdater",
    "OrderBookProducer",
    "OrderBookChannel",
    "OrderBookTickerProducer",
    "OrderBookTickerChannel",
    "OrderBookManager",
    "OrderBookUpdaterSimulator",
    "MarkPriceUpdaterSimulator",
    "MarkPriceProducer",
    "MarkPriceChannel",
    "PricesManager",
    "calculate_mark_price_from_recent_trade_prices",
    "MarkPriceUpdater",
    "PriceEventsManager",
    "RecentTradeProducer",
    "RecentTradeChannel",
    "LiquidationsProducer",
    "LiquidationsChannel",
    "RecentTradesManager",
    "RecentTradeUpdater",
    "TickerManager",
    "TickerUpdater",
    "TickerProducer",
    "TickerChannel",
    "MiniTickerProducer",
    "MiniTickerChannel",
    "TickerUpdaterSimulator",
    "MarginContract",
    "FutureContract",
    "ExchangeSymbolsData",
    "ExchangeSymbolData",
]



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport octobot_trading.util as util


cdef class PriceEventsManager(util.Initializable):
    cdef object logger

    cdef list events
    cdef list _last_recent_prices

    cpdef void reset(self)
    cpdef void handle_recent_trades(self, list recent_trades)
    cpdef void handle_price(self, object price, double timestamp)
    cpdef object new_event(self, object price, double timestamp, bint trigger_above, bint allow_instant_fill=*) # return asyncio.Event
    cpdef object remove_event(self, object event_to_remove) # object is an asyncio.Event
    cpdef void clear_recent_prices(self)

    cdef bint _is_triggered_by_last_recent_prices(self, object price, double timestamp, bint trigger_above)
    cdef void _add_recent_price(self, object price, double timestamp)
    cdef object _remove_and_set_event(self, object event_to_set) # return to propagate errors
    cdef object _remove_event(self, object event_to_remove) # object is an asyncio.Event
    cdef list _check_events(self, object price, double timestamp)

cdef tuple _new_price_event(object price, double timestamp, bint trigger_above)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library
import octobot_commons.display as commons_display
import octobot_commons.enums as commons_enums
import octobot_commons.databases as commons_databases
import octobot_commons.logging as commons_logging
import octobot_commons.errors as commons_errors

import octobot_trading.util as util
import octobot_trading.storage.trades_storage as trades_storage
import octobot_trading.storage.transactions_storage as transactions_storage
import octobot_trading.storage.candles_storage as candles_storage
import octobot_trading.storage.portfolio_storage as portfolio_storage


class StorageManager(util.Initializable):
    def __init__(self, exchange_manager):
        super().__init__()
        self.logger = commons_logging.get_logger(self.__class__.__name__)
        self.exchange_manager = exchange_manager
        self.trades_storage = None
        self.orders_storage = None
        self.transactions_storage = None
        self.portfolio_storage = None
        self.candles_storage = None

    async def initialize_impl(self):
        await commons_databases.RunDatabasesProvider.instance().get_run_databases_identifier(
            self.exchange_manager.bot_id
        ).initialize(
            exchange=self.exchange_manager.exchange_name
        )
        for storage in self._storages():
            try:
                await storage.start()
            except Exception as e:
                self.logger.exception(e, True, f"Error when initializing {storage}: {e}")

    async def stop(self):
        for storage in self._storages():
            await storage.stop()
        self.exchange_manager = None
        self.trades_storage = self.orders_storage = self.transactions_storage = \
            self.portfolio_storage = self.candles_storage = None

    async def store_history(self):
        if self.exchange_manager is None:
            raise commons_errors.MissingExchangeDataError("This exchange storage has been stopped")
        for storage in self._historical_storages():
            await storage.store_history()

    def _historical_storages(self):
        for storage in self._storages():
            if storage.is_historical:
                yield storage

    def _storages(self):
        return (
            self.trades_storage or self._trade_storage_factory(),
            self.transactions_storage or self._transaction_storage_factory(),
            self.candles_storage or self._candles_storage_factory(),
            self.portfolio_storage or self._portfolio_storage_factory(),
        )

    def _trade_storage_factory(self):
        self.trades_storage = trades_storage.TradesStorage(
            self.exchange_manager,
            commons_display.PlotSettings(
                chart=commons_enums.PlotCharts.MAIN_CHART.value,
                x_multiplier=1000,
                mode="markers",
                kind="scattergl",
            )
        )
        return self.trades_storage

    def _transaction_storage_factory(self):
        self.transactions_storage = transactions_storage.TransactionsStorage(
            self.exchange_manager,
            commons_display.PlotSettings(
                chart=commons_enums.PlotCharts.MAIN_CHART.value,
                x_multiplier=1000,
                mode="markers",
                kind="scattergl",
                y_data=None,
            )
        )
        return self.transactions_storage

    def _candles_storage_factory(self):
        self.candles_storage = candles_storage.CandlesStorage(
            self.exchange_manager,
            commons_display.PlotSettings(
                chart=commons_enums.PlotCharts.MAIN_CHART.value,
            )
        )
        return self.candles_storage

    def _portfolio_storage_factory(self):
        self.portfolio_storage = portfolio_storage.PortfolioStorage(
            self.exchange_manager,
            commons_display.PlotSettings()
        )
        return self.portfolio_storage



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

cpdef object create_order_from_raw(object trader,
                                   dict raw_order)

cpdef object create_order_instance_from_raw(object trader,
                                            dict raw_order,
                                            bint force_open= *)

cpdef object create_order_from_type(object trader,
                                    object order_type,
                                    object side= *)

cpdef object create_order_instance(object trader,
                                   object order_type,
                                   str symbol,
                                   object current_price,
                                   object quantity,
                                   object price= *,
                                   object stop_price= *,
                                   object status= *,
                                   str order_id= *,
                                   object filled_price= *,
                                   object average_price= *,
                                   object quantity_filled= *,
                                   object total_cost= *,
                                   double timestamp= *,
                                   object side= *,
                                   object fees_currency_side=*,
                                   object group=*,
                                   object tag=*,
                                   object reduce_only=*,
                                   str quantity_currency=*,
                                   bint close_position=*)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.


from octobot_trading.exchange_data.order_book.channel cimport order_book
from octobot_trading.exchange_data.order_book.channel.order_book cimport (
    OrderBookProducer,
    OrderBookChannel,
    OrderBookTickerProducer,
    OrderBookTickerChannel,
)

from octobot_trading.exchange_data.order_book.channel cimport order_book_updater
from octobot_trading.exchange_data.order_book.channel.order_book_updater cimport (
    OrderBookUpdater,
)

from octobot_trading.exchange_data.order_book.channel cimport (
    order_book_updater_simulator,
)
from octobot_trading.exchange_data.order_book.channel.order_book_updater_simulator cimport (
    OrderBookUpdaterSimulator,
)

__all__ = [
    "OrderBookUpdater",
    "OrderBookProducer",
    "OrderBookChannel",
    "OrderBookTickerProducer",
    "OrderBookTickerChannel",
    "OrderBookUpdaterSimulator",
]
